"""
Contains types for .pyi files.
Import from this at your own risk.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    MutableMapping,
    Optional,
    Tuple,
    TypedDict,
    Union,
)

from rich_click.utils import CommandGroupDict, OptionGroupDict


if TYPE_CHECKING:
    import sys

    from rich.align import AlignMethod
    from rich.box import Box
    from rich.console import Console, JustifyMethod
    from rich.padding import PaddingDimensions
    from rich.style import StyleType
    from rich.text import Text, TextType

    from rich_click.rich_click_theme import RichClickTheme
    from rich_click.rich_help_configuration import (
        CommandColumnType,
        CommandHelpSectionType,
        OptionColumnType,
        OptionHelpSectionType,
        RichHelpConfiguration,
    )

    if sys.version_info < (3, 11):
        from typing_extensions import NotRequired
    else:
        from typing import NotRequired


class RichContextSettingsDict(TypedDict):
    obj: NotRequired["Any | None"]
    auto_envvar_prefix: NotRequired[Optional[str]]
    default_map: NotRequired[MutableMapping[str, Any] | None]
    terminal_width: NotRequired[Optional[int]]
    max_content_width: NotRequired[Optional[int]]
    resilient_parsing: NotRequired[bool]
    allow_extra_args: NotRequired[Optional[bool]]
    allow_interspersed_args: NotRequired[Optional[bool]]
    ignore_unknown_options: NotRequired[Optional[bool]]
    help_option_names: NotRequired[Optional[List[str]]]
    token_normalize_func: NotRequired[Callable[[str], str] | None]
    color: NotRequired[Optional[bool]]
    show_default: NotRequired[Optional[bool]]
    rich_console: NotRequired[Optional["Console"]]
    rich_help_config: NotRequired[Optional[Union[Mapping[str, Any], RichHelpConfiguration]]]
    export_console_as: NotRequired[Optional[Literal["html", "svg", "text"]]]
    errors_in_output_format: NotRequired[Optional[bool]]
    help_to_stderr: NotRequired[Optional[bool]]


class TableKwargs(TypedDict):
    title: NotRequired[Optional["TextType"]]
    caption: NotRequired[Optional["TextType"]]
    width: NotRequired[Optional[int]]
    min_width: NotRequired[Optional[int]]
    box: NotRequired[Optional[Union[str, "Box"]]]
    safe_box: NotRequired[Optional[bool]]
    padding: NotRequired["PaddingDimensions"]
    collapse_padding: NotRequired[bool]
    pad_edge: NotRequired[bool]
    expand: NotRequired[bool]
    show_header: NotRequired[bool]
    show_footer: NotRequired[bool]
    show_edge: NotRequired[bool]
    show_lines: NotRequired[bool]
    leading: NotRequired[int]
    style: NotRequired["StyleType"]
    row_styles: NotRequired[Optional[Iterable["StyleType"]]]
    header_style: NotRequired[Optional["StyleType"]]
    footer_style: NotRequired[Optional["StyleType"]]
    border_style: NotRequired[Optional["StyleType"]]
    title_style: NotRequired[Optional["StyleType"]]
    caption_style: NotRequired[Optional["StyleType"]]
    title_justify: NotRequired["JustifyMethod"]
    caption_justify: NotRequired["JustifyMethod"]
    highlight: NotRequired[bool]


class PanelKwargs(TypedDict):
    box: NotRequired[Union[str, "Box"]]
    title: NotRequired[Optional["TextType"]]
    title_align: NotRequired["AlignMethod"]
    subtitle: NotRequired[Optional["TextType"]]
    subtitle_align: NotRequired["AlignMethod"]
    safe_box: NotRequired[Optional[bool]]
    expand: NotRequired[bool]
    style: NotRequired["StyleType"]
    border_style: NotRequired["StyleType"]
    width: NotRequired[Optional[int]]
    height: NotRequired[Optional[int]]
    padding: NotRequired["PaddingDimensions"]
    highlight: NotRequired[bool]


class RichHelpConfigurationDict(TypedDict):
    """Typed dict for rich_config() kwargs."""

    theme: NotRequired[Optional[Union[str, "RichClickTheme"]]]
    enable_theme_env_var: NotRequired[bool]
    style_option: NotRequired["StyleType"]
    style_option_negative: NotRequired[Optional["StyleType"]]
    style_argument: NotRequired["StyleType"]
    style_command: NotRequired["StyleType"]
    style_command_aliases: NotRequired["StyleType"]
    style_switch: NotRequired["StyleType"]
    style_switch_negative: NotRequired[Optional["StyleType"]]
    style_metavar: NotRequired["StyleType"]
    style_metavar_append: NotRequired["StyleType"]
    style_metavar_separator: NotRequired["StyleType"]
    style_range_append: NotRequired["StyleType"]
    style_header_text: NotRequired["StyleType"]
    style_epilog_text: NotRequired["StyleType"]
    style_footer_text: NotRequired["StyleType"]
    style_usage: NotRequired["StyleType"]
    style_usage_command: NotRequired["StyleType"]
    style_usage_separator: NotRequired["StyleType"]
    style_deprecated: NotRequired["StyleType"]
    style_helptext_first_line: NotRequired["StyleType"]
    style_helptext: NotRequired["StyleType"]
    style_helptext_aliases: NotRequired[Optional["StyleType"]]
    style_option_help: NotRequired["StyleType"]
    style_command_help: NotRequired["StyleType"]
    style_option_default: NotRequired["StyleType"]
    style_option_envvar: NotRequired["StyleType"]
    style_required_short: NotRequired["StyleType"]
    style_required_long: NotRequired["StyleType"]
    style_options_panel_border: NotRequired["StyleType"]
    style_options_panel_box: NotRequired[Optional[Union[str, "Box"]]]
    style_options_panel_help_style: NotRequired["StyleType"]
    style_options_panel_title_style: NotRequired["StyleType"]
    style_options_panel_padding: NotRequired["PaddingDimensions"]
    style_options_panel_style: NotRequired["StyleType"]
    align_options_panel: NotRequired["AlignMethod"]
    style_options_table_show_lines: NotRequired[bool]
    style_options_table_leading: NotRequired[int]
    style_options_table_pad_edge: NotRequired[bool]
    style_options_table_padding: NotRequired["PaddingDimensions"]
    style_options_table_expand: NotRequired[bool]
    style_options_table_box: NotRequired[Optional[Union[str, "Box"]]]
    style_options_table_row_styles: NotRequired[Optional[List["StyleType"]]]
    style_options_table_border_style: NotRequired[Optional["StyleType"]]
    style_commands_panel_border: NotRequired["StyleType"]
    panel_inline_help_in_title: NotRequired[bool]
    panel_inline_help_delimiter: NotRequired[str]
    style_commands_panel_box: NotRequired[Optional[Union[str, "Box"]]]
    style_commands_panel_help_style: NotRequired["StyleType"]
    style_commands_panel_title_style: NotRequired["StyleType"]
    style_commands_panel_padding: NotRequired["PaddingDimensions"]
    style_commands_panel_style: NotRequired["StyleType"]
    align_commands_panel: NotRequired[AlignMethod]
    style_commands_table_show_lines: NotRequired[bool]
    style_commands_table_leading: NotRequired[int]
    style_commands_table_pad_edge: NotRequired[bool]
    style_commands_table_padding: NotRequired["PaddingDimensions"]
    style_commands_table_expand: NotRequired[bool]
    style_commands_table_box: NotRequired[Optional[Union[str, "Box"]]]
    style_commands_table_row_styles: NotRequired[Optional[List["StyleType"]]]
    style_commands_table_border_style: NotRequired[Optional["StyleType"]]
    style_commands_table_column_width_ratio: NotRequired[Optional[Union[Tuple[None, None], Tuple[int, int]]]]
    style_errors_panel_border: NotRequired["StyleType"]
    style_errors_panel_box: NotRequired[Optional[Union[str, "Box"]]]
    align_errors_panel: NotRequired[AlignMethod]
    style_errors_suggestion: NotRequired[Optional["StyleType"]]
    style_errors_suggestion_command: NotRequired[Optional["StyleType"]]
    style_padding_errors: NotRequired["StyleType"]
    style_aborted: NotRequired["StyleType"]
    style_padding_usage: NotRequired["StyleType"]
    style_padding_helptext: NotRequired["StyleType"]
    style_padding_epilog: NotRequired["StyleType"]

    panel_title_padding: NotRequired[int]
    width: NotRequired[Optional[int]]
    max_width: NotRequired[Optional[int]]
    color_system: NotRequired[Optional[Literal["auto", "standard", "256", "truecolor", "windows"]]]
    force_terminal: NotRequired[Optional[bool]]
    options_table_column_types: NotRequired[List["OptionColumnType"]]
    commands_table_column_types: NotRequired[List["CommandColumnType"]]
    options_table_help_sections: NotRequired[List["OptionHelpSectionType"]]
    commands_table_help_sections: NotRequired[List["CommandHelpSectionType"]]

    header_text: NotRequired[Optional[Union[str, "Text"]]]
    footer_text: NotRequired[Optional[Union[str, "Text"]]]
    panel_title_string: NotRequired[str]
    deprecated_string: NotRequired[str]
    deprecated_with_reason_string: NotRequired[str]
    default_string: NotRequired[str]
    envvar_string: NotRequired[str]
    required_short_string: NotRequired[str]
    required_long_string: NotRequired[str]
    range_string: NotRequired[str]
    append_metavars_help_string: NotRequired[str]
    append_range_help_string: NotRequired[str]
    helptext_aliases_string: NotRequired[str]
    arguments_panel_title: NotRequired[str]
    options_panel_title: NotRequired[str]
    commands_panel_title: NotRequired[str]
    errors_panel_title: NotRequired[str]
    delimiter_comma: NotRequired[str]
    delimiter_slash: NotRequired[str]
    errors_suggestion: NotRequired[Optional[Union[str, Text]]]
    errors_epilogue: NotRequired[Optional[Union[str, Text]]]
    aborted_text: NotRequired[str]
    padding_header_text: NotRequired["PaddingDimensions"]
    padding_helptext: NotRequired["PaddingDimensions"]
    padding_helptext_deprecated: NotRequired["PaddingDimensions"]
    padding_helptext_first_line: NotRequired["PaddingDimensions"]
    padding_usage: NotRequired["PaddingDimensions"]
    padding_epilog: NotRequired["PaddingDimensions"]
    padding_footer_text: NotRequired["PaddingDimensions"]
    padding_errors_panel: NotRequired["PaddingDimensions"]
    padding_errors_suggestion: NotRequired["PaddingDimensions"]
    padding_errors_epilogue: NotRequired["PaddingDimensions"]

    # Behaviours
    show_arguments: NotRequired[Optional[bool]]
    show_metavars_column: NotRequired[Optional[bool]]
    commands_before_options: NotRequired[bool]
    append_metavars_help: NotRequired[Optional[bool]]
    group_arguments_options: NotRequired[bool]
    option_envvar_first: NotRequired[Optional[bool]]
    text_markup: NotRequired[Literal["ansi", "rich", "markdown", None]]
    text_kwargs: NotRequired[Optional[Dict[str, Any]]]
    text_emojis: NotRequired[bool]
    text_paragraph_linebreaks: NotRequired[Optional[Literal["\n", "\n\n"]]]
    use_markdown: NotRequired[Optional[bool]]
    use_markdown_emoji: NotRequired[Optional[bool]]
    use_rich_markup: NotRequired[Optional[bool]]
    command_groups: NotRequired[Dict[str, List[CommandGroupDict]]]
    option_groups: NotRequired[Dict[str, List[OptionGroupDict]]]
    use_click_short_help: NotRequired[bool]
    helptext_show_aliases: NotRequired[bool]
    highlighter_patterns: NotRequired[List[str]]
    legacy_windows: NotRequired[Optional[bool]]



from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    MutableMapping,
    NoReturn,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

import click

# Group, Command, and CommandCollection need to be imported directly,
# or else rich_click.cli.patch() causes a recursion error.
from click import Group
from rich.console import Console

from rich_click._internal_types import RichContextSettingsDict
from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_help_formatter import RichHelpFormatter
from rich_click.rich_help_rendering import RichPanelRow
from rich_click.rich_panel import RichCommandPanel, RichPanel

_AnyCallable = Callable[..., Any]
C = TypeVar("C", bound=click.Command)
G = TypeVar("G", bound=click.Group)

# TLDR: if a subcommand overrides one of the methods called by `RichCommand.format_help`,
# then the text won't render properly. The fix is to not rely on the composability of the API,
# and to instead force everything to use RichCommand's methods.
OVERRIDES_GUARD: bool = False

class RichCommand(click.Command):

    context_class: Type[RichContext] = RichContext
    _formatter: Optional[RichHelpFormatter] = None
    panels: List[RichPanel[Any, Any]]
    panel: Optional[str]
    aliases: Iterable[str]

    def __init__(
        self,
        *args: Any,
        aliases: Optional[Iterable[str]] = None,
        panels: Optional[List[RichPanel[Any, Any]]] = None,
        panel: Optional[str] = None,
        name: str | None,
        context_settings: MutableMapping[str, Any] | None = None,
        callback: Callable[..., Any] | None = None,
        params: list[click.Parameter] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        add_help_option: bool = True,
        no_args_is_help: bool = False,
        hidden: bool = False,
        deprecated: bool | str = False,
    ) -> None: ...
    @property
    def console(self) -> Optional["Console"]: ...
    @property
    def help_config(self) -> Optional[RichHelpConfiguration]: ...
    def _generate_rich_help_config(self) -> RichHelpConfiguration: ...
    def _error_formatter(self) -> RichHelpFormatter: ...
    @overload
    def main(
        self,
        args: Sequence[str] | None = None,
        prog_name: str | None = None,
        complete_var: str | None = None,
        standalone_mode: Literal[True] = True,
        **extra: Any,
    ) -> NoReturn: ...
    @overload
    def main(
        self,
        args: Sequence[str] | None = None,
        prog_name: str | None = None,
        complete_var: str | None = None,
        standalone_mode: bool = ...,
        **extra: Any,
    ) -> Any: ...
    def main(
        self,
        args: Optional[Sequence[str]] = None,
        prog_name: Optional[str] = None,
        complete_var: Optional[str] = None,
        standalone_mode: bool = True,
        windows_expand_args: bool = True,
        **extra: Any,
    ) -> Any: ...
    def format_help(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def format_help_text(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def format_options(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def format_epilog(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def get_help_option(self, ctx: RichContext) -> Union[click.Option, None]: ...
    def get_rich_table_row(
        self,
        ctx: RichContext,
        formatter: RichHelpFormatter,
        panel: Optional[RichCommandPanel] = None,
    ) -> RichPanelRow: ...
    def add_panel(self, panel: "RichPanel[Any, Any]") -> None: ...
    def add_command_to_panel(
        self,
        command_name: str,
        panel_name: Union[str, Iterable[str]],
    ) -> None: ...

class RichGroup(RichCommand, click.Group):
    """
    Richly formatted click Group.

    Inherits click.Group and overrides help and error methods
    to print richly formatted output.
    """

    command_class: Optional[Type[RichCommand]] = RichCommand
    group_class: Optional[Union[Type[Group], Type[type]]] = type
    _alias_mapping: Dict[str, str]
    _panel_command_mapping: Dict[str, List[str]]

    def __init__(
        self,
        panels: Optional[List["RichPanel[Any, Any]"]] = None,
        aliases: Optional[Iterable[str]] = None,
        name: str | None = None,
        commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = None,
        invoke_without_command: bool = False,
        no_args_is_help: bool | None = None,
        subcommand_metavar: str | None = None,
        chain: bool = False,
        result_callback: Callable[..., Any] | None = None,
        context_settings: MutableMapping[str, Any] | None = None,
        callback: Callable[..., Any] | None = None,
        params: list[click.Parameter] | None = None,
        help: str | None = None,
        epilog: str | None = None,
        short_help: str | None = None,
        options_metavar: str | None = "[OPTIONS]",
        add_help_option: bool = True,
        hidden: bool = False,
        deprecated: bool | str = False,
    ) -> None: ...
    def format_commands(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def format_help(self, ctx: RichContext, formatter: RichHelpFormatter) -> None: ...
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    # variant: no call, directly as decorator for a function.
    @overload
    def command(self, name: _AnyCallable) -> RichCommand: ...
    @overload
    def command(
        self,
        name: Optional[str] = ...,
        *,
        cls: Type[C],
        context_settings: RichContextSettingsDict,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        no_args_is_help: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], C]: ...
    @overload
    def command(
        self,
        name: Optional[str] = ...,
        *,
        cls: None,
        context_settings: RichContextSettingsDict,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        no_args_is_help: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], RichCommand]: ...
    @overload
    def command(
        self,
        name: Optional[str] = ...,
        *,
        cls: Type[C],
        context_settings: MutableMapping[str, Any] | None = ...,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        no_args_is_help: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], C]: ...
    @overload
    def command(
        self,
        name: Optional[str] = None,
        *,
        cls: None,
        context_settings: MutableMapping[str, Any] | None = ...,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        no_args_is_help: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
        **attrs: Any,
    ) -> Callable[[_AnyCallable], RichCommand]: ...

    # variant: with positional name and with positional or keyword cls argument:
    # @command(namearg, CommandCls, ...) or @command(namearg, cls=CommandCls, ...)
    @overload
    def command(
        self,
        name: Optional[str],
        cls: Type[C],
        **attrs: Any,
    ) -> Callable[[_AnyCallable], C]: ...

    # variant: name omitted, cls _must_ be a keyword argument, @command(cls=CommandCls, ...)
    @overload
    def command(
        self,
        name: None = None,
        *,
        cls: Type[C],
        **attrs: Any,
    ) -> Callable[[_AnyCallable], C]: ...

    # variant: with optional string name, no cls argument provided.
    @overload
    def command(
        self, name: Optional[str] = ..., cls: None = None, **attrs: Any
    ) -> Callable[[_AnyCallable], RichCommand]: ...
    def command(
        self, name: Optional[str] = None, *, cls: Optional[Type[C]] = None, **kwargs: Any
    ) -> Callable[[_AnyCallable], Union[click.Command, C]]: ...
    @overload
    def group(self, name: _AnyCallable) -> RichGroup: ...
    @overload
    def group(
        self,
        name: Optional[str] = None,
        *,
        cls: Type[G],
        commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
        invoke_without_command: bool = ...,
        no_args_is_help: bool | None = ...,
        subcommand_metavar: str | None = ...,
        chain: bool = ...,
        result_callback: Callable[..., Any] | None = ...,
        context_settings: RichContextSettingsDict,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], G]: ...
    @overload
    def group(
        self,
        name: Optional[str] = ...,
        *,
        cls: None,
        commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
        invoke_without_command: bool = ...,
        no_args_is_help: bool | None = ...,
        subcommand_metavar: str | None = ...,
        chain: bool = ...,
        result_callback: Callable[..., Any] | None = ...,
        context_settings: MutableMapping[str, Any] | None = ...,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], RichGroup]: ...
    @overload
    def group(
        self,
        name: Optional[str] = None,
        *,
        cls: Type[G],
        commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
        invoke_without_command: bool = ...,
        no_args_is_help: bool | None = ...,
        subcommand_metavar: str | None = ...,
        chain: bool = ...,
        result_callback: Callable[..., Any] | None = ...,
        context_settings: MutableMapping[str, Any] | None = ...,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], G]: ...
    @overload
    def group(
        self,
        name: Optional[str] = ...,
        *,
        cls: None,
        commands: MutableMapping[str, click.Command] | Sequence[click.Command] | None = ...,
        invoke_without_command: bool = ...,
        no_args_is_help: bool | None = ...,
        subcommand_metavar: str | None = ...,
        chain: bool = ...,
        result_callback: Callable[..., Any] | None = ...,
        context_settings: RichContextSettingsDict,
        callback: Callable[..., Any] | None = ...,
        params: list[click.Parameter] | None = ...,
        help: str | None = ...,
        epilog: str | None = ...,
        short_help: str | None = ...,
        options_metavar: str | None = ...,
        add_help_option: bool = ...,
        hidden: bool = ...,
        deprecated: bool | str = ...,
        aliases: Optional[Iterable[str]] = ...,
        panels: Optional[List[RichPanel[Any, Any]]] = ...,
        panel: Optional[str] = ...,
    ) -> Callable[[_AnyCallable], RichGroup]: ...

    # variant: with positional name and with positional or keyword cls argument:
    # @group(namearg, GroupCls, ...) or @group(namearg, cls=GroupCls, ...)
    @overload
    def group(
        self,
        name: Optional[str],
        cls: Type[G],
        **attrs: Any,
    ) -> Callable[[_AnyCallable], G]: ...

    # variant: name omitted, cls _must_ be a keyword argument, @group(cmd=GroupCls, ...)
    @overload
    def group(
        self,
        name: None = None,
        *,
        cls: Type[G],
        **attrs: Any,
    ) -> Callable[[_AnyCallable], G]: ...

    # variant: with optional string name, no cls argument provided.
    @overload
    def group(
        self, name: Optional[str] = ..., cls: None = None, **attrs: Any
    ) -> Callable[[_AnyCallable], RichGroup]: ...
    def group(
        self,
        name: Union[str, _AnyCallable, None] = None,
        cls: Optional[Type[G]] = None,
        **attrs: Any,
    ) -> Union[click.Group, Callable[[_AnyCallable], Union[RichGroup, G]]]: ...
    def get_command(self, ctx: RichContext, cmd_name: str) -> Optional[click.Command]: ...
    def add_command(
        self,
        cmd: click.Command,
        name: str | None = None,
        aliases: Optional[Iterable[str]] = None,
        panel: Optional[str] = None,
    ) -> None: ...
    def _handle_extras_add_command(
        self,
        cmd: click.Command,
        name: Optional[str] = None,
        aliases: Optional[Iterable[str]] = None,
        panel: Optional[Union[str, List[str]]] = None,
    ) -> None: ...

class RichMultiCommand(RichGroup, click.CommandCollection):
    pass

class RichCommandCollection(RichGroup, click.CommandCollection):
    pass



"""The command line interface."""

from __future__ import annotations

import os
import sys
import warnings
from contextlib import contextmanager
from functools import wraps
from gettext import gettext
from importlib import (
    import_module,
    metadata,  # type: ignore[import,unused-ignore]
)
from typing import Any, Dict, Generator, List, Literal, Optional, Tuple

import click
from click.core import ParameterSource

from rich_click.decorators import argument as _rich_argument
from rich_click.decorators import command as _rich_command
from rich_click.decorators import option as _rich_option
from rich_click.decorators import option_panel, pass_context
from rich_click.decorators import version_option as _rich_version_option
from rich_click.patch import patch as _patch
from rich_click.rich_click_theme import COLORS, FORMATS, RichClickThemeNotFound, get_theme
from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import RichHelpConfiguration


DISABLE_WARNINGS_NOTE = (
    "[red]You can pass [b]--suppress-warnings[/b] to the rich-click CLI"
    " or set the [b]RICH_CLICK_CLI_SUPPRESS_WARNINGS[/b] environment variable"
    " to disable this message.[/red]"
)


def entry_points(*, group: str) -> "metadata.EntryPoints":
    """entry_points function that is compatible with Python 3.7+."""
    if sys.version_info >= (3, 10):
        return metadata.entry_points(group=group)

    epg = metadata.entry_points()

    return epg.get(group, [])


@wraps(_patch)
def patch(*args: Any, **kwargs: Any) -> None:  # noqa: D103
    import warnings

    warnings.warn(
        "`rich_click.cli.patch()` has moved to `rich_click.patch.patch()`."
        " Importing `patch()` from `rich_click.cli` is deprecated; please import from `rich_click.patch` instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    return _patch(*args, **kwargs)


class _RichHelpConfigurationParamType(click.ParamType):
    name = "JSON"

    def __repr__(self) -> str:
        return "JSON"

    def convert(
        self,
        value: Optional[str],
        param: Optional[click.Parameter],
        ctx: Optional[click.Context],
    ) -> Optional[Dict[str, Any]]:
        if value is None:
            return value
        else:
            try:
                import json

                if value.startswith("@"):
                    with open(value[1:], "r") as f:
                        data = json.load(f)
                else:
                    data = json.loads(value)
                if not isinstance(data, dict):
                    raise ValueError("--rich-config needs to be a JSON.")
                return data
            except Exception as e:
                # In normal circumstances, a bad arg to a CLI doesn't
                # prevent the help text from rendering.
                if ctx is not None and ctx.params.get("show_help", False):
                    print(ctx.get_help())
                    ctx.exit()
                else:
                    raise e


def _get_module_path_and_function_name(script: str, suppress_warnings: bool) -> Tuple[str, str]:
    _selected: List[str] = []
    module_path = ""
    function_name = ""

    for s in entry_points(group="console_scripts"):
        if script == s.name:
            if not _selected:
                module_path, function_name = s.value.split(":", 1)
                if " [" in function_name and function_name.endswith("]"):
                    function_name = function_name.split(" [")[0]
            if suppress_warnings:
                break
            if s.value not in _selected:
                _selected.append(s.value)

    if len(_selected) > 1 and not suppress_warnings:
        # This is an extremely rare edge case that comes up when the user sets the PYTHONPATH themselves.
        if script in sys.argv:
            _args = sys.argv.copy()
            _args[_args.index(script)] = f"{module_path}:{function_name}"
        else:
            _args = ["rich-click", f"{module_path}:{function_name}"]

        import rich

        rich.print(
            f"[red]WARNING: Multiple entry_points correspond with script '{script}': {_selected!r}."
            "\nThis can happen when an 'egg-info' directory exists, you're using a virtualenv,"
            " and you have set a custom PYTHONPATH."
            f"\n\nThe selected script is '{module_path}:{function_name}', which is being executed now."
            "\n\nIt is safer and recommended that you specify the MODULE:CLICK_COMMAND"
            f" ('{module_path}:{function_name}') instead of the script ('{script}'), like this:[/]"
            f"\n\n>>> {' '.join(_args)}"
            f"\n\n{DISABLE_WARNINGS_NOTE}",
        )

    if ":" in script and not module_path:
        # the path to a function was passed
        module_path, function_name = script.split(":", 1)

    if not module_path:
        module_path = script

    return module_path, function_name


def list_themes(ctx: RichContext, param: click.Parameter, value: bool) -> None:
    """Print all themes."""
    if value:
        from rich import box
        from rich.console import Console
        from rich.containers import Renderables
        from rich.padding import Padding
        from rich.panel import Panel
        from rich.table import Table
        from rich.text import Text

        formatter = ctx.make_formatter()

        console = Console(
            color_system=formatter.config.color_system,
            force_terminal=formatter.config.force_terminal,
            width=formatter.config.width,
            legacy_windows=formatter.config.legacy_windows,
        )

        found = False
        c, f = None, None

        selected = os.getenv("RICH_CLICK_THEME")
        if selected is not None:
            if selected.startswith("{"):
                import json

                try:
                    data = json.loads(selected)
                    _theme = data.get("theme")
                except Exception:
                    _theme = None
            else:
                _theme = selected
            if _theme:
                try:
                    get_theme(_theme)
                except RichClickThemeNotFound:
                    found = False
                else:
                    found = True
                    if "-" in _theme:
                        c, f, *_ = _theme.split("-")
                    elif _theme in FORMATS:
                        f = _theme
                        found = True
                    elif _theme in COLORS:
                        c = _theme
                        found = True

        if selected is None:
            selected_text = "[red bold]RICH_CLICK_THEME[/] is not set"
        elif not found:
            selected_text = f"[red bold]RICH_CLICK_THEME[/]=[red]{selected!r}[/] is not a valid theme."
        else:
            selected_text = (
                f"[blue bold]RICH_CLICK_THEME[/blue bold]=[blue]{selected!r}[/blue]"
                f"\n\n[b]Colors:[/b] {c or '[dim](null)[/]'}"
                f"\n[b]Format:[/b] {f or '[dim](null)[/]'}"
            )

        colors = Table(
            "",
            "Name",
            "Colors",
            "Description",
            padding=(0, 1),
            border_style="dim",
            box=box.SIMPLE_HEAD,
            show_lines=True,
            expand=True,
        )
        colors.columns[0].max_width = 1
        colors.columns[1].style = "bold"
        for r in COLORS.values():
            if r.hidden:
                continue
            selected_row = c == r.name or (c is None and r.name == "default")
            colors.add_row(
                "*" if selected_row else "",
                r.name,
                " ".join([f"[{_}]▓▓[/]" for _ in r.primary_colors[:3]]),
                r.description,
            )

        formats = Table(
            "",
            "Name",
            "Description",
            padding=(0, 1),
            border_style="dim",
            box=box.SIMPLE_HEAD,
            show_lines=True,
            expand=True,
        )
        formats.columns[1].style = "bold"
        for r in FORMATS.values():
            selected_row = f == r.name or (f is None and r.name == "box")
            formats.add_row("*" if selected_row else "", r.name, r.description)

        how_to_use = Padding(
            Text("\n\n", overflow="fold").join(
                [
                    Text.from_markup("[b]Themes[/b] are an easy way to style a rich-click CLI."),
                    Text.from_markup(
                        "As an end-user of CLIs, you can set the [blue b]RICH_CLICK_THEME=[/blue b] env var"
                        " to style all rich-click CLIs you use."
                    ),
                    Text.from_markup(
                        "As a developer, you can add a theme to your CLI with one line of code:"
                        " [green b]@click.rich_config({'theme': 'name'})[/green b]."
                    ),
                    Text.from_markup(
                        "Themes consist of [b]Color Palettes[/b] and [b]Formats[/b], which can be mixed and matched:"
                        " The name of a full theme has the following schema: [red]{color_palette}-{format}[/red]."
                    ),
                    Text.from_markup(
                        "For example, the [b]forest-slim[/b] theme uses the [b]forest[/b]"
                        " color palette and the [b]slim[/b] format."
                    ),
                ]
            ),
            pad=1,
            expand=False,
        )

        console.print(
            Panel(
                Renderables(
                    [
                        Panel(how_to_use, title="How to use", width=100),
                        Panel(Padding(selected_text, pad=1), title="Local user setting", width=100),
                        Panel(colors, title="Color Palettes"),
                        Panel(formats, title="Formats"),
                    ]
                ),
                expand=False,
                box=box.SIMPLE,
            )
        )
        ctx.exit(0)


@_rich_command(
    "rich-click",
    context_settings={"allow_interspersed_args": False},
    add_help_option=False,
)
@_rich_argument(
    "script_and_args",
    nargs=-1,
    metavar="SCRIPT | MODULE[:CLICK_COMMAND] ...",
    help="The script you want to run. If it's a Click CLI and you are rendering help text;"
    " then the help text will render"
    " [#FF6B6B bold]r[/][#FF8E53 bold]i[/][#FFB347 bold]c[/][#4ECDC4 bold]h[/]"
    "[#45B7D1 bold]l[/][#74B9FF bold]y[/]. Otherwise, the script will run normally.",
)
@_rich_option(
    "--theme",
    "-t",
    help="Set the theme to render the CLI with.",
    metavar="THEME",
)
@_rich_option(
    "--rich-config",
    "-c",
    type=_RichHelpConfigurationParamType(),
    help="Keyword arguments to pass into the [de]RichHelpConfiguration()[/] used"
    " to render the help text of the command. You can pass either a JSON directly, or a file"
    " prefixed with `@` (for example: '@rich_config.json'). Note that the --rich-config"
    " option is also used to render this help text you're reading right now!",
)
@_rich_option(
    "--output",
    "-o",
    type=click.Choice(["html", "svg", "text"], case_sensitive=False),
    help="Optionally render help text as HTML or SVG or plain text. By default, help text is rendered normally.",
)
@_rich_option(
    "--errors-in-output-format",
    is_flag=True,
    panel="Advanced Options",
    help="If set, forces the CLI to render CLI error messages"
    " in the format specified by the --output option."
    " By default, error messages render normally, i.e. they are not converted to html or svg.",
)
@_rich_option(
    "--suppress-warnings/--do-not-suppress-warnings",
    is_flag=True,
    default=False,
    envvar="RICH_CLICK_CLI_SUPPRESS_WARNINGS",
    show_envvar=True,
    panel="Advanced Options",
    help="Suppress warnings when there are conflicting entry_points. This situation is extremely rare.",
)
@_rich_option(
    "--patch-rich-click/--no-patch-rich-click",
    is_flag=True,
    default=True,
    panel="Advanced Options",
    help="If set, patch [reverse][option][b]rich_click.Command[/][/][/],"
    " not just [reverse][option][b]click.Command[/][/][/].",
)
@_rich_option(
    "--themes",
    help="List all available themes and exit.",
    panel="Extra",
    callback=list_themes,  # type: ignore[arg-type,unused-ignore]
    expose_value=False,
    is_flag=True,
)
@_rich_version_option(panel="Extra")
@_rich_option(
    # The rich-click CLI uses a special implementation of --help,
    # which is aware of the --rich-config object.
    "--help",
    "-h",
    "show_help",
    is_eager=True,
    is_flag=True,
    help=gettext("Show this message and exit."),
    panel="Extra",
    # callback=help_callback
)
@option_panel("Options")
@option_panel("Advanced Options")
@option_panel("Extra")
@pass_context
def main(
    ctx: RichContext,
    script_and_args: Tuple[str, ...],
    theme: str,
    output: Literal[None, "html", "svg"],
    errors_in_output_format: bool,
    suppress_warnings: bool,
    patch_rich_click: bool,
    rich_config: Optional[Dict[str, Any]],
    show_help: bool,
) -> None:
    """
    The [link=https://github.com/ewels/rich-click]rich-click[/] CLI provides
    [#FF6B6B bold]r[/][#FF8E53 bold]i[/][#FFB347 bold]c[/][#4ECDC4 bold]h[/][#45B7D1 bold]l[/][#74B9FF bold]y[/]
    formatted help output from any
    tool using [link=https://click.palletsprojects.com/]click[/], formatted with
    [link=https://github.com/Textualize/rich]rich[/].

    Full docs here: [link=https://ewels.github.io/rich-click/latest/documentation/rich_click_cli/]\
https://ewels.github.io/rich-click/latest/documentation/rich_click_cli/[/]

    The rich-click command line tool can be prepended before any Python package
    using native click to provide attractive richified click help output.

    For example, if you have a package called [argument]my_package[/] that uses click,
    you can run:

    >>> [command]rich-click[/] [argument]my_package[/] [option]--help[/]

    When not rendering help text, the provided command will run normally,
    so it is safe to replace calls to the tool with [command]rich-click[/] in front, e.g.:

    >>> [command]rich-click[/] [argument]my_package[/] [argument]cmd[/] [option]--foo[/] 3
    """  # noqa: D401
    try:
        if rich_config:
            if theme:
                rich_config.setdefault("theme", theme)
                import rich_click.rich_click as rc

                rc._THEME_FROM_CLI = theme
            cfg = RichHelpConfiguration.load_from_globals(**rich_config)
        elif theme:
            import rich_click.rich_click as rc

            rc._THEME_FROM_CLI = theme
            cfg = RichHelpConfiguration.load_from_globals(theme=theme)
        else:
            cfg = RichHelpConfiguration.load_from_globals()
    except RichClickThemeNotFound as e:
        raise click.ClickException(e.args[0] if e.args else "Theme not found")

    if (show_help or not script_and_args) and not ctx.resilient_parsing:
        cfg.use_markdown = False
        cfg.use_rich_markup = True
        cfg.text_markup = "rich"
        if cfg.show_arguments is None:
            cfg.show_arguments = False
        ctx.help_config = cfg
        print(ctx.get_help())
        if not show_help and not script_and_args:
            ctx.exit(2)
        ctx.exit(0)

    @contextmanager
    def patch_ctx() -> Generator[None, None, None]:
        if ctx.get_parameter_source("typer") != ParameterSource.ENVIRONMENT:
            try:
                with warnings.catch_warnings():
                    warnings.filterwarnings("error", category=ImportWarning)
                    yield
            except ImportWarning as e:
                if not suppress_warnings and any("Typer" in a for a in e.args):
                    import rich

                    rich.print(
                        f"[red]WARNING: Attempted to patch Typer, but Typer is not installed.[/red]"
                        f"\n\n{DISABLE_WARNINGS_NOTE}"
                    )
        else:
            yield

    # patch click before importing the program function
    with patch_ctx():
        _patch(rich_config=cfg, patch_rich_click=patch_rich_click)

    script, *args = script_and_args

    # PYTHONPATH can change output of entry_points(group="console_scripts") in rare cases,
    # so we want to run the whole search with this.
    _sys_path = sys.path.copy()
    sys.path.append(os.path.abspath("."))
    try:
        module_path, function_name = _get_module_path_and_function_name(script, suppress_warnings)
    finally:
        sys.path = _sys_path

    if function_name:

        def function() -> None:
            try:
                module = import_module(module_path)
            except ModuleNotFoundError:
                try:
                    # Import can fail if module is relative to root dir
                    # and PYTHONPATH does not include ".".
                    sys.path.append(os.path.abspath("."))
                    module = import_module(module_path)
                except ModuleNotFoundError as e:
                    raise click.ClickException(e.args[0] if e.args else "Unknown error")
            try:
                f = getattr(module, function_name)
            except AttributeError:
                raise click.ClickException(f"Module '{module_path}' has no attribute '{function_name}'.")
            f()

    elif module_path.endswith(".py"):

        def function() -> None:
            import runpy

            try:
                runpy.run_path(module_path, run_name="__main__")
            except ImportError as e:
                raise click.ClickException(e.args[0] if e.args else "Unknown error")

    else:

        def function() -> None:
            import runpy

            try:
                runpy.run_module(module_path, run_name="__main__")
            except ImportError:
                # Import can fail if module is relative to root dir
                # and PYTHONPATH does not include ".".
                sys.path.append(os.path.abspath("."))
                try:
                    runpy.run_module(module_path, run_name="__main__")
                except ImportError as e:
                    raise click.ClickException(e.args[0] if e.args else "Unknown error")

    prog = module_path.split(".", 1)[0]
    sys.argv = [prog, *args]

    RichContext.export_console_as = ctx.export_console_as = output
    RichContext.errors_in_output_format = ctx.errors_in_output_format = errors_in_output_format
    function()



from __future__ import annotations

import inspect
import re
from enum import Enum
from gettext import gettext
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Literal, Optional, Tuple, Union, overload

import click
from rich.align import Align
from rich.columns import Columns
from rich.console import RenderableType, group
from rich.highlighter import RegexHighlighter
from rich.jupyter import JupyterMixin
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from rich_click._click_types_cache import Argument, Command, Group, Option
from rich_click._compat_click import (
    CLICK_IS_BEFORE_VERSION_82,
    CLICK_IS_VERSION_80,
)
from rich_click.rich_context import RichContext
from rich_click.rich_help_formatter import RichHelpFormatter
from rich_click.rich_parameter import RichParameter


if TYPE_CHECKING:
    from rich.markdown import Markdown

    from rich_click.rich_help_configuration import CommandColumnType, OptionColumnType, OptionHelpSectionType
    from rich_click.rich_panel import RichCommandPanel, RichOptionPanel


RichPanelRow = List[Optional[RenderableType]]


class RichClickRichPanel(Panel):
    """
    A console renderable that draws a border around its contents.

    This is a patched version of rich.panel.Panel that has additional features useful
    for rendering help text with rich-click.
    """

    def __init__(self, *args: Any, title_padding: int = 1, **kwargs: Any) -> None:
        """
        Create RichClickRichPanel instance.

        Args:
        ----
            *args: Args that get passed to rich.panel.Panel.
            title_padding: Controls padding on panel title.
            **kwargs: Kwargs that get passed to rich.panel.Panel.

        """
        super().__init__(*args, **kwargs)
        self.title_padding = title_padding

    @property
    def _title(self) -> Optional[Text]:
        if self.title:
            title_text = Text.from_markup(self.title) if isinstance(self.title, str) else self.title.copy()
            title_text.end = ""
            title_text.plain = title_text.plain.replace("\n", " ")
            title_text.no_wrap = True
            title_text.expand_tabs()
            # Ensure underlines are handled beautifully;
            # title_text.pad() expands underlines, whereas this method does not.
            if self.title_padding == 0:
                return title_text
            return Text("").join([Text(" " * self.title_padding), title_text, Text(" " * self.title_padding)])
        return None


@group()
def _get_help_text(
    obj: Union[Command, Group], formatter: RichHelpFormatter
) -> Iterable[Union[Padding, "Markdown", Text]]:
    """
    Build primary help text for a click command or group.
    Returns the prose help text for a command or group, rendered either as a
    Rich Text object or as Markdown.
    If the command is marked as depreciated, the depreciated string will be prepended.

    Args:
    ----
        obj (click.Command or click.Group): Command or group to build help text for.
        formatter: formatter object.

    Yields:
    ------
        Text or Markdown: Multiple styled objects (depreciated, usage)

    """
    if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(obj.help, str)
    config = formatter.config
    # Prepend deprecated status
    if obj.deprecated:
        if isinstance(obj.deprecated, str):
            yield Padding(
                Text.from_markup(
                    formatter.config.deprecated_with_reason_string.format(obj.deprecated.replace("[", r"\[")),
                    style=config.style_deprecated,
                ),
                formatter.config.padding_helptext_deprecated,
                style=formatter.config.style_padding_helptext,
            )
        else:
            yield Padding(
                Text.from_markup(config.deprecated_string, style=config.style_deprecated),
                formatter.config.padding_helptext_deprecated,
                style=formatter.config.style_padding_helptext,
            )

    # Fetch and dedent the help text
    help_text = inspect.cleandoc(obj.help or "")

    # Trim off anything that comes after \f on its own line
    help_text = help_text.partition("\f")[0]

    # Get the first paragraph
    first_line = help_text.split("\n\n")[0]
    # Remove single linebreaks
    if not config.text_markup == "markdown":
        if not first_line.startswith("\b"):
            first_line = first_line.replace("\n", " ")
    yield Padding(
        formatter.rich_text(first_line.strip(), formatter.config.style_helptext_first_line),
        formatter.config.padding_helptext_first_line,
        style=formatter.config.style_padding_helptext,
    )
    # Get remaining lines, remove single line breaks and format as dim
    remaining_paragraphs = help_text.split("\n\n")[1:]

    use_markdown = formatter.config.text_markup == "markdown"
    if formatter.config.text_paragraph_linebreaks is None:
        if use_markdown:
            lb = "\n\n"
        else:
            lb = "\n"
    else:
        lb = formatter.config.text_paragraph_linebreaks

    if len(remaining_paragraphs) > 0:
        if not use_markdown:
            # Remove single linebreaks
            help_text_buf = []
            for para in remaining_paragraphs:
                if para.startswith("\b"):
                    help_text_buf.append("{}\n".format(para.strip("\b\n")))
                    help_text_buf.append(lb)
                else:
                    _continuation = False
                    _list = False
                    _first = True
                    for p in para.split("\n"):
                        if any(p.startswith(_) for _ in ["* ", "- "]):
                            _continuation = False
                            if not _first:
                                help_text_buf.append("\n")
                            _list = True
                            help_text_buf.append(p)
                        elif _list and p.startswith("  "):
                            help_text_buf.append(" ")
                            help_text_buf.append(p.lstrip())
                        elif any(p.startswith(_) for _ in ["    ", "> "]):
                            _continuation = False
                            _list = False
                            if not _first:
                                help_text_buf.append("\n")
                            help_text_buf.append(p)
                        elif _continuation:
                            help_text_buf.append(" ")
                            help_text_buf.append(p)
                        else:
                            _list = False
                            help_text_buf.append(p)
                            _continuation = True
                        _first = False
                help_text_buf.append(lb)
            # Join back together
            remaining_lines = "".join(help_text_buf)
        else:
            # Join with double linebreaks if markdown
            remaining_lines = lb.join(remaining_paragraphs)
        yield formatter.rich_text(remaining_lines, formatter.config.style_helptext)
    if getattr(obj, "aliases", None) and formatter.config.helptext_show_aliases:
        yield Text.from_markup(
            formatter.config.helptext_aliases_string.format(", ".join(obj.aliases)),  # type: ignore[attr-defined]
            style=(
                formatter.config.style_helptext_aliases
                if formatter.config.style_helptext_aliases is not None
                else formatter.config.style_helptext
            ),
        )


def _get_deprecated_text(
    deprecated: Union[bool, str],
    formatter: RichHelpFormatter,
) -> Text:
    if isinstance(deprecated, str):
        s = formatter.config.deprecated_with_reason_string.format(deprecated.replace("[", r"\["))
    else:
        s = formatter.config.deprecated_string
    return Text.from_markup(s, style=formatter.config.style_deprecated)


def _get_parameter_env_var(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Optional[Text]:
    if not getattr(param, "show_envvar", None):
        return None

    envvar = getattr(param, "envvar", None)

    # https://github.com/pallets/click/blob/0aec1168ac591e159baf6f61026d6ae322c53aaf/src/click/core.py#L2720-L2726
    if envvar is None:
        if (
            getattr(param, "allow_from_autoenv", None)
            and getattr(ctx, "auto_envvar_prefix", None) is not None
            and param.name is not None
        ):
            envvar = f"{ctx.auto_envvar_prefix}_{param.name.upper()}"
    if envvar is not None:
        envvar = ", ".join(envvar) if isinstance(envvar, list) else envvar

    if envvar is not None:
        return Text.from_markup(
            formatter.config.envvar_string.format(envvar), style=formatter.config.style_option_envvar
        )
    return None


def _get_parameter_deprecated(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Optional[Text]:
    if not getattr(param, "deprecated", None):
        return None
    return _get_deprecated_text(getattr(param, "deprecated"), formatter)


def _get_parameter_help(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Optional[Union["Markdown", Text]]:
    base_help_txt = getattr(param, "help", None)
    if not base_help_txt:
        return None

    if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(param, click.Option)
        assert hasattr(param, "help")
        assert isinstance(param.help, str)

    paragraphs = base_help_txt.split("\n\n")

    # Remove single linebreaks
    if not formatter.config.use_markdown and not formatter.config.text_markup == "markdown":
        paragraphs = [
            x.replace("\n", " ").strip() if not x.startswith("\b") else "{}\n".format(x.strip("\b\n"))
            for x in paragraphs
        ]
    help_text = "\n".join(paragraphs).strip()

    # `Deprecated` is included in base help text; remove it here.
    if getattr(param, "deprecated", None):
        if isinstance(getattr(param, "deprecated"), str):
            help_text = re.sub(r"\(DEPRECATED: .*?\)$", "", help_text)
        else:
            help_text = re.sub(r"\(DEPRECATED\)$", "", help_text)

    if getattr(param, "help_style", None) is None:
        style = formatter.config.style_option_help
    else:
        style = param.help_style  # type: ignore[attr-defined]
    return formatter.rich_text(help_text, style)


@overload
def _get_parameter_range(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    mode: Literal["metavar_append"],
) -> Optional[str]: ...


@overload
def _get_parameter_range(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    mode: Literal["metavar_column", "help"],
) -> Optional[Text]: ...


def _get_parameter_range(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    mode: Literal["metavar_append", "metavar_column", "help"],
) -> Optional[Union[Text, str]]:
    # Range - from
    # https://github.com/pallets/click/blob/c63c70dabd3f86ca68678b4f00951f78f52d0270/src/click/core.py#L2698-L2706  # noqa: E501
    # skip count with default range type
    if (
        hasattr(param, "count")
        and isinstance(param.type, click.types._NumberRangeBase)
        and not (param.count and param.type.min == 0 and param.type.max is None)
    ):
        range_str = param.type._describe_range()
        if range_str:
            if mode == "metavar_append":
                return range_str
            elif mode == "metavar_column":
                metavar_str = formatter.config.range_string.format(range_str)
                return Text.from_markup(metavar_str, style=formatter.config.style_metavar)
            elif mode == "help":
                metavar_str = formatter.config.append_range_help_string.format(range_str)
                return Text.from_markup(
                    metavar_str,
                    style=(
                        formatter.config.style_range_append
                        if formatter.config.style_range_append is not None
                        else formatter.config.style_metavar_append
                    ),
                )
            else:
                raise ValueError("Bad mode selected")
    return None


def _get_parameter_metavar(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    append: bool = True,
    show_range: bool = False,
) -> Optional[Text]:
    metavar_str = param.make_metavar() if CLICK_IS_BEFORE_VERSION_82 else param.make_metavar(ctx)  # type: ignore
    # Do it ourselves if this is a positional argument
    if isinstance(param, Argument) and param.name is not None and re.match(rf"\[?{param.name.upper()}]?", metavar_str):
        metavar_str = param.type.name.upper()
    # Attach metavar if param is a positional argument, or if it is a non boolean and non flag option
    if isinstance(param, Argument) or (metavar_str != "BOOLEAN" and hasattr(param, "is_flag") and not param.is_flag):
        metavar_str = metavar_str.replace("[", "").replace("]", "")

        if show_range:
            range_txt = _get_parameter_range(param, ctx, formatter, mode="metavar_append")
            if range_txt:
                metavar_str += " " + range_txt

        return Text.from_markup(
            formatter.config.append_metavars_help_string.format(metavar_str),
            style=formatter.config.style_metavar_append if append else formatter.config.style_metavar,
            overflow="fold",
        )
    return None


def _get_parameter_help_metavar_col(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    show_range: bool = True,
) -> Optional[Text]:
    # Column for a metavar, if we have one
    metavar = Text(style=formatter.config.style_metavar, overflow="fold")
    metavar_str = param.make_metavar() if CLICK_IS_BEFORE_VERSION_82 else param.make_metavar(ctx)  # type: ignore

    if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(param.name, str)
        assert isinstance(param, Option)

    # Do it ourselves if this is a positional argument
    if isinstance(param, Argument) and re.match(rf"\[?{param.name.upper()}]?", metavar_str):
        metavar_str = param.type.name.upper()

    # Attach metavar if param is a positional argument, or if it is a non-boolean and non flag option
    if isinstance(param, Argument) or (metavar_str != "BOOLEAN" and not getattr(param, "is_flag", None)):
        metavar.append(metavar_str)

    # Range - from
    # https://github.com/pallets/click/blob/c63c70dabd3f86ca68678b4f00951f78f52d0270/src/click/core.py#L2698-L2706  # noqa: E501
    try:
        # skip count with default range type
        if isinstance(param.type, click.types._NumberRangeBase) and not (
            param.count and param.type.min == 0 and param.type.max is None
        ):
            range_str = param.type._describe_range()
            if show_range and range_str:
                metavar.append(" " + formatter.config.range_string.format(range_str))
    except AttributeError:
        # click.types._NumberRangeBase is only in Click 8x onwards
        pass

    # Highlighter to make [ | ] and <> dim
    class MetavarHighlighter(RegexHighlighter):
        highlights = [
            r"(^|\s)(?P<metavar_sep>(\[|<))",
            r"(?P<metavar_sep>\|)",
            r"(?P<metavar_sep>(\]|>)$)",
        ]

    if "".join(metavar._text) == "":
        return None

    metavar_highlighter = MetavarHighlighter()
    return metavar_highlighter(metavar)


def _get_parameter_help_opt(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Tuple[
    Optional[Text],
    Optional[Text],
    Optional[Text],
    Optional[Text],
    Optional[Text],
]:
    # This may seem convoluted to return tuples with widths instead of just renderables,
    # but there are two things we want to do that are impossible otherwise:
    # 1. Prevent splitting of options. If we just use Text() and allow wrapping,
    #    then something like --foo/--bar may split in the middle of an option
    #    instead of at the "/".
    # 2. Make it so the table doesn't expand. We solve the first problem by using
    #    Columns(), but the issue is there is no way to prevent the expansion of
    #    the outer table column other than to explicitly set the width.
    # Attempting to solve both of those problems simultaneously leads to this mess.

    opt_long_primary = []
    opt_short_primary = []
    opt_long_secondary = []
    opt_short_secondary = []

    for opt in param.opts:
        if isinstance(param, Argument):
            opt_long_primary.append(opt.upper())
        elif "--" in opt:
            opt_long_primary.append(opt)
        else:
            opt_short_primary.append(opt)
    for opt in param.secondary_opts:
        if isinstance(param, Argument):
            opt_long_secondary.append(opt.upper())
        elif "--" in opt:
            opt_long_secondary.append(opt)
        else:
            opt_short_secondary.append(opt)

    from rich.text import Text

    # opt_long_primary_len = len("".join(opt_long_primary)) + len(opt_long_primary) - 1
    # opt_short_primary_len = len("".join(opt_short_primary)) + len(opt_short_primary) - 1
    # opt_long_secondary_len = len("".join(opt_long_secondary)) + len(opt_long_secondary) - 1
    # opt_short_secondary_len = len("".join(opt_short_secondary)) + len(opt_short_secondary) - 1

    primary_cols = []
    secondary_cols = []
    long_cols = []
    short_cols = []

    comma = Text(formatter.config.delimiter_comma, style=formatter.config.style_option_help)
    slash = Text(formatter.config.delimiter_slash, style=formatter.config.style_option_help)

    for o in opt_short_primary:
        oh = Text(o.strip(), style=formatter.config.style_switch)
        primary_cols.append(oh)
        primary_cols.append(comma)
        short_cols.append(oh)
        short_cols.append(comma)

    for o in opt_long_primary:
        oh = Text(o.strip(), style=formatter.config.style_option)
        primary_cols.append(oh)
        primary_cols.append(comma)
        long_cols.append(oh)
        long_cols.append(comma)

    if opt_short_secondary:
        short_cols = short_cols[:-1]
        short_cols.append(slash)
        for o in opt_short_secondary:
            oh = Text(
                o.strip(),
                style=(
                    formatter.config.style_switch_negative
                    if formatter.config.style_switch_negative is not None
                    else formatter.config.style_switch
                ),
            )
            secondary_cols.append(oh)
            secondary_cols.append(comma)
            short_cols.append(oh)
            short_cols.append(comma)

    if opt_long_secondary:
        long_cols = long_cols[:-1]
        long_cols.append(slash)
        for o in opt_long_secondary:
            oh = Text(
                o.strip(),
                style=(
                    formatter.config.style_option_negative
                    if formatter.config.style_option_negative is not None
                    else formatter.config.style_option
                ),
            )
            secondary_cols.append(oh)
            secondary_cols.append(comma)
            long_cols.append(oh)
            long_cols.append(comma)

    all_cols = primary_cols
    if secondary_cols:
        all_cols = [*primary_cols[:-1], slash, *secondary_cols]

    def _renderable(cols: List[Text]) -> Optional[Text]:
        if not cols:
            return None
        if len(cols) == 1:
            return cols[0]
        # Todo - make the text fold at the slashes without adding whitespace.
        #   this is very tricky.
        t = Text("", overflow="fold").join(cols)
        return t

    primary_final = _renderable(primary_cols[:-1])
    secondary_final = _renderable(secondary_cols[:-1])
    long_final = _renderable(long_cols[:-1])
    short_final = _renderable(short_cols[:-1])
    all_final = _renderable(all_cols[:-1])

    return primary_final, secondary_final, long_final, short_final, all_final


def _get_parameter_default(
    param: Union[click.Argument, click.Option, RichParameter], ctx: RichContext, formatter: RichHelpFormatter
) -> Optional[Text]:

    if not hasattr(param, "show_default"):
        return None

    show_default = False
    show_default_is_str = False

    resilient = ctx.resilient_parsing
    ctx.resilient_parsing = True
    try:
        default_value = param.get_default(ctx, call=False)
    finally:
        ctx.resilient_parsing = resilient

    notset: Tuple[Any, ...]
    try:
        # try/except because it's unclear whether later versions of click will change this.
        if not CLICK_IS_BEFORE_VERSION_82:
            from click.core import UNSET  # type: ignore[attr-defined,unused-ignore]

            notset = (UNSET, None)
        else:
            notset = (None,)
    except ImportError:
        notset = (None,)

    if (not CLICK_IS_VERSION_80 and param.show_default is not None) or param.show_default:
        if isinstance(param.show_default, str):
            show_default_is_str = show_default = True
        else:
            show_default = param.show_default
    elif ctx.show_default is not None:
        show_default = ctx.show_default

    default_string: Optional[str] = None

    if show_default_is_str or (show_default and (default_value not in notset)):
        if show_default_is_str:
            default_string = f"({param.show_default})"
        elif isinstance(default_value, (list, tuple)):
            default_string = ", ".join(str(d) for d in default_value)
        elif isinstance(default_value, Enum):
            default_string = str(default_value.value)
        elif inspect.isfunction(default_value):
            default_string = gettext("(dynamic)")
        elif hasattr(param, "is_bool_flag") and param.is_bool_flag and param.secondary_opts:
            # For boolean flags that have distinct True/False opts,
            # use the opt without prefix instead of the value.
            opt = (param.opts if default_value else param.secondary_opts)[0]
            first = opt[:1]
            if first.isalnum():
                default_string = opt
            if opt[1:2] == first:
                default_string = opt[2:]
            else:
                default_string = opt[1:]
        elif hasattr(param, "is_bool_flag") and param.is_bool_flag and not param.secondary_opts and not default_value:
            if CLICK_IS_VERSION_80:
                default_string = str(param.default)
            else:
                default_string = ""
        elif default_value == "":
            default_string = '""'
        else:
            default_string = str(default_value)

    if default_string:
        return Text.from_markup(
            formatter.config.default_string.format(default_string.replace("[", r"\[")),
            style=formatter.config.style_option_default,
        )
    return None


def _get_parameter_required(
    param: Union[click.Argument, click.Option, RichParameter], ctx: RichContext, formatter: RichHelpFormatter
) -> Optional[Text]:
    if param.required:
        return Text.from_markup(formatter.config.required_long_string, style=formatter.config.style_required_long)
    return None


def _get_parameter_help_required_short(
    param: Union[click.Argument, click.Option, RichParameter], ctx: RichContext, formatter: RichHelpFormatter
) -> Optional[Text]:
    if param.required:
        return Text(formatter.config.required_short_string, style=formatter.config.style_required_short)
    return None


def get_help_parameter(
    param: Union[click.Argument, click.Option, RichParameter], ctx: RichContext, formatter: RichHelpFormatter
) -> Columns:
    """
    Build primary help text for a click option or argument.
    Returns the prose help text for an option or argument, rendered either
    as a Rich Text object or as Markdown.
    Additional elements are appended to show the default and required status if applicable.

    Args:
    ----
        param (click.Argument or click.Option): Parameter to build help text for.
        ctx (click.Context): Click Context object.
        formatter (RichHelpFormatter): formatter object.

    Returns:
    -------
        Columns: A columns element with multiple styled objects (help, default, required)

    """
    if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(param.name, str)

    section_callbacks: Dict["OptionHelpSectionType", Callable[..., Any]] = {
        "help": _get_parameter_help,
        "required": _get_parameter_required,
        "envvar": _get_parameter_env_var,
        "default": _get_parameter_default,
        "metavar": lambda param, ctx, formatter: _get_parameter_metavar(param, ctx, formatter, show_range=True),
        "metavar_short": lambda param, ctx, formatter: _get_parameter_metavar(param, ctx, formatter, show_range=False),
        "range": lambda param, ctx, formatter: _get_parameter_range(param, ctx, formatter, mode="help"),
        "deprecated": _get_parameter_deprecated,
    }

    sections: List[Optional[RenderableType]] = []
    for sec in formatter.config.options_table_help_sections:
        sections.append(section_callbacks[sec](param, ctx, formatter))

    # Use Columns - this allows us to group different renderable types
    # (Text, Markdown) onto a single line.
    if formatter.config.text_markup == "markdown" or not all([isinstance(i, Text) or i is None for i in sections]):
        return Columns([i for i in sections if i])
    else:
        # Weird but necessary--
        # in order to keep things flush, the last column of the table must be of type Columns().
        # (We are assuming here 'help' is always last, which is not necessarily the case of course.)
        # In a 2.0 of rich-click we will try to do something nicer than this.
        # But in 1.x all other solutions would be too breaking.
        return Columns(
            [
                Text(" ", overflow="fold", style=formatter.config.style_option_help).join(
                    [i for i in sections if i]  # type: ignore[misc]
                )
            ]
        )


def get_parameter_rich_table_row(
    param: Union[click.Argument, click.Option, RichParameter],
    ctx: RichContext,
    formatter: RichHelpFormatter,
    panel: Optional["RichOptionPanel"],
) -> RichPanelRow:
    """Create a row for the rich table corresponding with this parameter."""
    # Short and long form
    column_types: List["OptionColumnType"]
    if panel is None:
        column_types = formatter.config.options_table_column_types
    else:
        column_types = panel.column_types or formatter.config.options_table_column_types

    opt_long_strs = []
    opt_short_strs = []
    for idx, opt in enumerate(param.opts):
        opt_str = opt
        secondary = None
        try:
            secondary = param.secondary_opts[idx]
        except IndexError:
            pass

        if isinstance(param, Argument):
            opt_long_strs.append(Text.from_markup(opt_str.upper(), style=formatter.config.style_option))
        elif "--" in opt:
            if secondary:
                opt_long_strs.append(
                    Text("/", style=formatter.config.style_option_help).join(
                        [
                            Text(opt_str, style=formatter.config.style_option),
                            Text(
                                secondary,
                                style=(
                                    formatter.config.style_option_negative
                                    if formatter.config.style_option_negative is not None
                                    else formatter.config.style_option
                                ),
                            ),
                        ]
                    )
                )
            else:
                opt_long_strs.append(Text.from_markup(opt_str, style=formatter.config.style_option))
        else:
            if secondary:
                opt_short_strs.append(
                    Text("/", style=formatter.config.style_option_help).join(
                        [
                            Text(opt_str, style=formatter.config.style_option),
                            Text(
                                secondary,
                                style=(
                                    formatter.config.style_option_negative
                                    if formatter.config.style_option_negative is not None
                                    else formatter.config.style_option
                                ),
                            ),
                        ]
                    )
                )
            else:
                opt_short_strs.append(Text.from_markup(opt_str, style=formatter.config.style_option))

    if TYPE_CHECKING:  # pragma: no cover
        assert isinstance(param.name, str)
        assert isinstance(param, Option)

    _primary, _secondary, _long, _short, _all = _get_parameter_help_opt(param, ctx, formatter)

    _metavar_padded = None
    if any(i in column_types for i in ["opt_all_metavar", "opt_long_metavar"]):
        _metavar_padded = _get_parameter_metavar(param, ctx, formatter, append=False, show_range=False)

    def _opt_all_metavar() -> Optional[RenderableType]:
        if _metavar_padded is None:
            return _all
        if _all is None:
            return _metavar_padded
        return Text(" ", style=formatter.config.style_option_help).join([_all, _metavar_padded])

    def _opt_long_metavar() -> Optional[RenderableType]:
        if _metavar_padded is None:
            return _long
        if _long is None:
            return _metavar_padded
        return Text(" ", style=formatter.config.style_option_help).join([_long, _metavar_padded])

    column_callbacks: Dict["OptionColumnType", Callable[..., Any]] = {
        "required": _get_parameter_help_required_short,
        "opt_long": lambda *args, **kwargs: _long,
        "opt_short": lambda *args, **kwargs: _short,
        "opt_primary": lambda *args, **kwargs: _primary,
        "opt_secondary": lambda *args, **kwargs: _secondary,
        "opt_all": lambda *args, **kwargs: _all,
        "metavar": _get_parameter_help_metavar_col,
        "metavar_short": lambda *args, **kwargs: _get_parameter_help_metavar_col(*args, **kwargs, show_range=False),  # type: ignore[misc]
        "opt_all_metavar": lambda *args, **kwargs: _opt_all_metavar(),
        "opt_long_metavar": lambda *args, **kwargs: _opt_long_metavar(),
        "help": lambda *args, **kwargs: (
            param.get_rich_help(ctx, formatter)
            if isinstance(param, RichParameter)
            else get_help_parameter(param, ctx, formatter)
        ),
        # "default": lambda *args, **kwargs: None,
        # "envvar": lambda *args, **kwargs: None,
    }

    cols: RichPanelRow = []
    for col in column_types:
        cols.append(column_callbacks[col](param, ctx, formatter))

    return cols


def _get_command_name_help(
    command: click.Command,
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Text:
    # We want to use the command name as it is registered in the group.
    # This resolves the situation where they do not match.
    # e.g. group.add_command(subcmd, name="foo")
    #
    # Note there may be extremely weird edge cases not covered here,
    # most notably, when a user registers the same command twice.
    # We do not care to solve this edge case.
    command_name = None
    if isinstance(ctx.command, Group):
        if command.name not in ctx.command.commands:
            for k, v in ctx.command.commands.items():
                if command is v:
                    command_name = k
                    break
    if command_name is None:
        command_name = command.name or ""
    return Text(command_name, style=formatter.config.style_command)


def _get_command_aliases_help(
    command: click.Command,
    ctx: RichContext,
    formatter: RichHelpFormatter,
    include_name: bool = False,
) -> Optional[Text]:
    aliases = getattr(command, "aliases", None)
    if aliases:
        txt_list = []
        comma = Text(formatter.config.delimiter_comma, style=formatter.config.style_command_help)
        _last = len(aliases) - 1
        for idx, alias in enumerate(aliases):
            txt_list.append(Text(alias, style=formatter.config.style_command_aliases))
            if idx != _last:
                txt_list.append(comma)
        if include_name:
            cmd_name_txt = _get_command_name_help(command, ctx, formatter)
            return Text("", overflow="ellipsis").join([cmd_name_txt, comma, *txt_list])
        else:
            return Text("").join(txt_list)
    elif include_name:
        return _get_command_name_help(command, ctx, formatter)
    return None


def get_command_rich_table_row(
    command: click.Command,
    ctx: RichContext,
    formatter: RichHelpFormatter,
    panel: Optional["RichCommandPanel"],
) -> RichPanelRow:
    """Create a row for the rich table corresponding with this command."""
    # todo
    column_types: List["CommandColumnType"]
    if panel is None:
        column_types = formatter.config.commands_table_column_types
    else:
        column_types = panel.column_types or formatter.config.commands_table_column_types

    column_callbacks: Dict["CommandColumnType", Callable[..., Any]] = {
        "name": _get_command_name_help,
        "aliases": _get_command_aliases_help,
        "name_with_aliases": lambda *args, **kwargs: _get_command_aliases_help(*args, **kwargs, include_name=True),  # type: ignore[misc]
        "help": _get_command_help,
        # "short_help": ...,
    }

    cols: RichPanelRow = []
    for col in column_types:
        cols.append(column_callbacks[col](command, ctx, formatter))

    return cols


def _get_command_help(
    command: click.Command,
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> Union[Text, "Markdown", Columns]:
    """
    Build cli help text for a click group command.
    That is, when calling help on groups with multiple subcommands
    (not the main help text when calling the subcommand help).
    Returns the first paragraph of help text for a command, rendered either as a
    Rich Text object or as Markdown.
    Ignores single newlines as paragraph markers, looks for double only.

    Returns
    -------
        Text or Markdown: Styled object

    """
    if formatter.config.use_click_short_help:
        help_text = command.get_short_help_str()
    else:
        help_text = command.short_help or command.help or ""

    deprecated = command.deprecated

    paragraphs = inspect.cleandoc(help_text).split("\n\n")
    # Remove single linebreaks
    if not formatter.config.text_markup == "markdown" and not paragraphs[0].startswith("\b"):
        paragraphs[0] = paragraphs[0].replace("\n", " ")
    elif paragraphs[0].startswith("\b"):
        paragraphs[0] = paragraphs[0].replace("\b\n", "")
    help_text = paragraphs[0].strip()
    renderable: Union[Text, "Markdown", Columns]
    renderable = formatter.rich_text(help_text, formatter.config.style_command_help)
    if deprecated:
        dep_txt = _get_deprecated_text(
            deprecated=deprecated,
            formatter=formatter,
        )
        if isinstance(renderable, Text):
            renderable.append(" ")
            renderable.append(dep_txt)
        else:
            renderable = Columns([renderable, dep_txt])
    return renderable


def get_rich_usage(formatter: RichHelpFormatter, prog: str, args: str = "", prefix: Optional[str] = None) -> None:
    """Richly render usage text."""
    if prefix is None:
        prefix = "Usage:"

    config = formatter.config

    # Header text if we have it
    if config.header_text:
        formatter.write(
            Padding(
                formatter.rich_text(config.header_text, config.style_header_text),
                config.padding_header_text,
                style=formatter.config.style_padding_usage,
            ),
        )

    # Highlighter for options and arguments
    class UsageHighlighter(RegexHighlighter):
        highlights = [
            r"(?P<argument>\w+)",
        ]

    usage_highlighter = UsageHighlighter()

    # Print usage
    formatter.write(
        Padding(
            Columns(
                (
                    Text(prefix, style=config.style_usage),
                    Text(prog, style=config.style_usage_command),
                    usage_highlighter(Text(args, style=config.style_usage_separator)),
                )
            ),
            formatter.config.padding_usage,
            style=formatter.config.style_padding_usage,
        ),
    )


def get_rich_help_text(self: Command, ctx: RichContext, formatter: RichHelpFormatter) -> None:
    """Write rich help text to the formatter if it exists."""
    # Print command / group help if we have some
    if self.help or self.deprecated or getattr(self, "aliases", None):
        # Print with some padding
        formatter.write(
            Padding(
                Align(_get_help_text(self, formatter), pad=False),
                formatter.config.padding_helptext,
                style=formatter.config.style_padding_helptext,
            )
        )


def get_rich_epilog(
    self: Command,
    ctx: RichContext,
    formatter: RichHelpFormatter,
) -> None:
    """Richly render a click Command's epilog if it exists."""
    if self.epilog:
        # Remove single linebreaks, replace double with single
        lines = self.epilog.split("\n\n")
        if isinstance(self.epilog, JupyterMixin):  # Handles Text and Markdown
            epilog = self.epilog
        else:
            epilog = "\n".join([x.replace("\n", " ").strip() for x in lines])  # type: ignore[assignment]
            epilog = formatter.rich_text(epilog, formatter.config.style_epilog_text)  # type: ignore[assignment]
        formatter.write(
            Padding(
                Align(epilog, pad=False), formatter.config.padding_epilog, style=formatter.config.style_padding_epilog
            )
        )

    # Footer text if we have it
    if formatter.config.footer_text:
        formatter.write(
            Padding(
                formatter.rich_text(formatter.config.footer_text, formatter.config.style_footer_text),
                formatter.config.padding_footer_text,
                style=formatter.config.style_padding_epilog,
            )
        )


def rich_format_error(
    self: click.ClickException, formatter: RichHelpFormatter, export_console_as: Literal[None, "html", "svg"] = None
) -> None:
    """
    Print richly formatted click errors.

    Called by custom exception handler to print richly formatted click errors.
    Mimics original click.ClickException.echo() function but with rich formatting.

    Args:
    ----
        self (click.ClickException): Click exception to format.
        formatter: formatter object.
        export_console_as: If set, outputs error message as HTML or SVG.

    """
    config = formatter.config
    # Print usage
    if getattr(self, "ctx", None) is not None:
        if TYPE_CHECKING:  # pragma: no cover
            assert hasattr(self, "ctx")
        self.ctx.command.format_usage(self.ctx, formatter)
    if config.errors_suggestion:
        formatter.write(
            Padding(
                config.errors_suggestion,
                config.padding_errors_suggestion,
            ),
            style=config.style_errors_suggestion,
        )
    elif (
        config.errors_suggestion is None
        and getattr(self, "ctx", None) is not None
        and self.ctx.command.get_help_option(self.ctx) is not None  # type: ignore[attr-defined]
    ):
        cmd_path = self.ctx.command_path  # type: ignore[attr-defined]
        help_option = self.ctx.help_option_names[0]  # type: ignore[attr-defined]
        formatter.write(
            Padding(
                Text(" ").join(
                    (
                        Text("Try"),
                        Text(
                            f"'{cmd_path} {help_option}'",
                            style=(
                                config.style_errors_suggestion_command
                                if config.style_errors_suggestion_command is not None
                                else config.style_option
                            ),
                        ),
                        Text("for help"),
                    ),
                ),
                config.padding_errors_suggestion,
                style=config.style_padding_errors,
            ),
            style=(
                config.style_errors_suggestion if config.style_errors_suggestion is not None else config.style_helptext
            ),
        )

    # A major Python library using click (dbt-core) has its own exception
    # logic that subclasses ClickException, but does not use the message
    # attribute. Checking for the 'message' attribute works to make the
    # rich-click CLI compatible.
    if hasattr(self, "message"):

        from rich_click.rich_box import get_box

        formatter.write(
            Padding(
                Panel(
                    formatter.highlighter(self.format_message()),
                    border_style=config.style_errors_panel_border,
                    title=config.errors_panel_title,
                    title_align=config.align_errors_panel,
                    box=get_box(formatter.config.style_errors_panel_box or "ROUNDED"),
                ),
                config.padding_errors_panel,
                style=config.style_padding_errors,
            )
        )
    if config.errors_epilogue:
        formatter.write(Padding(config.errors_epilogue, config.padding_errors_epilogue))



from importlib import reload
from unittest.mock import patch

import pytest
from click.testing import CliRunner
from rich.console import Console

import rich_click
from rich_click._compat_click import CLICK_IS_BEFORE_VERSION_9X


@pytest.mark.skipif(not CLICK_IS_BEFORE_VERSION_9X, reason="Click 9 removes MultiCommand")
def test_deprecation_warning_on_multi_command() -> None:
    reload(rich_click)

    with pytest.warns(DeprecationWarning):
        from rich_click import RichMultiCommand  # noqa: F401


def test_deprecation_warning_on_highlighter(cli_runner: CliRunner) -> None:
    c = Console()

    @rich_click.rich_config({})
    @rich_click.pass_context
    def cli(ctx: rich_click.RichContext) -> None:
        if ctx.console is c:
            print("OK!")
        else:
            print("ERROR!")

    # console should be allowed to be first argument,
    with pytest.warns(DeprecationWarning):
        cli = rich_click.rich_config(c)(cli)  # type: ignore[arg-type]

    cli = rich_click.command()(cli)

    res = cli_runner.invoke(cli)
    assert res.stdout == "OK!\n"


def test_deprecation_cli_patch() -> None:

    with patch("rich_click.cli._patch") as mock_underlying_patch:
        with pytest.warns(DeprecationWarning, match=r"`rich_click\.cli\.patch\(\)` has moved.*"):
            import rich_click.cli

            rich_click.cli.patch(a=1, b=2)

    mock_underlying_patch.assert_called_once_with(a=1, b=2)


def test_deprecation_command_help_config() -> None:

    @rich_click.command()
    def cli1() -> None:
        pass

    with pytest.warns(DeprecationWarning):
        cfg1 = getattr(cli1, "help_config")
    assert cfg1 is None

    @rich_click.command()
    @rich_click.rich_config({})
    def cli2() -> None:
        pass

    with pytest.warns(DeprecationWarning):
        cfg2 = getattr(cli2, "help_config")
    assert isinstance(cfg2, rich_click.RichHelpConfiguration)


def test_deprecation_command_console() -> None:

    @rich_click.command()
    def cli1() -> None:
        pass

    with pytest.warns(DeprecationWarning):
        con1 = getattr(cli1, "console")
    assert con1 is None

    @rich_click.command()
    @rich_click.rich_config(console=Console())
    def cli2() -> None:
        pass

    with pytest.warns(DeprecationWarning):
        con2 = getattr(cli2, "console")
    assert isinstance(con2, Console)


def test_not_implemented_warnings_for_help_formatter() -> None:

    formatter = rich_click.RichHelpFormatter()

    with pytest.warns(RuntimeWarning):
        formatter.indent()
    with pytest.warns(RuntimeWarning):
        formatter.dedent()
    with pytest.warns(RuntimeWarning):
        formatter.write_heading("")
    with pytest.warns(RuntimeWarning):
        formatter.write_paragraph()
    with pytest.warns(RuntimeWarning):
        formatter.write_text("")
    with pytest.warns(RuntimeWarning):
        formatter.write_dl([])
    with pytest.warns(RuntimeWarning):
        with formatter.section(""):
            pass
    with pytest.warns(RuntimeWarning):
        with formatter.indentation():
            pass



# ruff: noqa: D101,D103,D401,E501
import json
import sys
from importlib.metadata import version
from pathlib import Path
from typing import Callable, List

import packaging.version
import pytest
from click.testing import CliRunner
from inline_snapshot import snapshot

from rich_click.cli import main
from tests.conftest import WriteScript, run_as_subprocess


@pytest.fixture
def simple_script(mock_script_writer: WriteScript) -> Path:
    return mock_script_writer(
        '''
        import click

        # Test if robust to subclassing
        class CustomClickCommand(click.Command):
            pass

        @click.command(cls=CustomClickCommand)
        def cli():
            """My help text"""
            print('Hello, world!')
        '''
    )


@pytest.mark.parametrize(
    "command",
    [
        ["mymodule:cli", "--help"],
        ["--", "mymodule:cli", "--help"],
    ],
)
def test_simple_rich_click_cli(simple_script: Path, command: List[str]) -> None:
    res = run_as_subprocess([sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"])
    assert res.returncode == 0
    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


@pytest.mark.parametrize(
    "command",
    [
        ["mymodule:cli"],
        ["--", "mymodule:cli"],
        ["mymodule:cli", "--"],
    ],
)
def test_simple_rich_click_cli_execute_command(simple_script: Path, cli_runner: CliRunner, command: List[str]) -> None:
    res = cli_runner.invoke(main, command)

    assert res.exit_code == 0
    assert res.stdout == "Hello, world!\n"

    subprocess_res = run_as_subprocess([sys.executable, "-m", "src.rich_click", *command])
    assert subprocess_res.returncode == 0

    assert subprocess_res.stdout.decode() == "Hello, world!\n"


def test_rich_click_cli_help() -> None:
    res = run_as_subprocess([sys.executable, "-m", "src.rich_click", "--help"])
    assert res.returncode == 0


def test_rich_click_cli_help_with_rich_config() -> None:
    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--rich-config", '{"style_option": "bold red"}', "--help"]
    )
    assert res.returncode == 0


def test_rich_click_cli_help_with_rich_config_from_file(tmp_path: Path) -> None:
    config_data = {"options_panel_title": "Custom Name"}
    config_file = tmp_path / "myconfig.json"
    config_file.write_text(json.dumps(config_data))

    res = run_as_subprocess([sys.executable, "-m", "src.rich_click", "--rich-config", f"@{config_file}", "--help"])
    assert res.returncode == 0
    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click [OPTIONS] SCRIPT | MODULE[:CLICK_COMMAND] ...                      \n\
                                                                                                    \n\
 The rich-click CLI provides richly formatted help output from any tool using click, formatted with \n\
 rich.                                                                                              \n\
 Full docs here: https://ewels.github.io/rich-click/latest/documentation/rich_click_cli/            \n\
 The rich-click command line tool can be prepended before any Python package using native click to  \n\
 provide attractive richified click help output.                                                    \n\
 For example, if you have a package called my_package that uses click, you can run:                 \n\
 >>> rich-click my_package --help                                                                   \n\
 When not rendering help text, the provided command will run normally, so it is safe to replace     \n\
 calls to the tool with rich-click in front, e.g.:                                                  \n\
 >>> rich-click my_package cmd --foo 3                                                              \n\
                                                                                                    \n\
╭─ Advanced Options ───────────────────────────────────────────────────────────────────────────────╮
│ --errors-in-output-format                       If set, forces the CLI to render CLI error       │
│                                                 messages in the format specified by the --output │
│                                                 option. By default, error messages render        │
│                                                 normally, i.e. they are not converted to html or │
│                                                 svg.                                             │
│ --suppress-warnings/--do-not-suppress-warnings  Suppress warnings when there are conflicting     │
│                                                 entry_points. This situation is extremely rare.  │
│                                                 [env var: RICH_CLICK_CLI_SUPPRESS_WARNINGS]      │
│ --patch-rich-click/--no-patch-rich-click        If set, patch rich_click.Command, not just       │
│                                                 click.Command.                                   │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
╭─ Extra ──────────────────────────────────────────────────────────────────────────────────────────╮
│ --themes       List all available themes and exit.                                               │
│ --version      Show the version and exit.                                                        │
│ --help     -h  Show this message and exit.                                                       │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
╭─ Custom Name ────────────────────────────────────────────────────────────────────────────────────╮
│ --theme        -t  THEME            Set the theme to render the CLI with.                        │
│ --rich-config  -c  JSON             Keyword arguments to pass into the RichHelpConfiguration()   │
│                                     used to render the help text of the command. You can pass    │
│                                     either a JSON directly, or a file prefixed with `@` (for     │
│                                     example: '@rich_config.json'). Note that the --rich-config   │
│                                     option is also used to render this help text you're reading  │
│                                     right now!                                                   │
│ --output       -o  [html|svg|text]  Optionally render help text as HTML or SVG or plain text. By │
│                                     default, help text is rendered normally.                     │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_rich_click_cli_help_with_bad_rich_config() -> None:
    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--rich-config", '{"bad", "json"}', "--help"],
    )
    assert res.returncode == 0


def test_rich_click_cli_help_with_bad_rich_config_v2() -> None:
    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--rich-config", "[]", "--help"],
    )
    assert res.returncode == 0


def test_custom_config_rich_click_cli(simple_script: Path) -> None:
    res = run_as_subprocess(
        [
            sys.executable,
            "-m",
            "src.rich_click",
            "--rich-config",
            '{"options_panel_title": "Custom Name"}',
            "mymodule:cli",
            "--help",
        ]
    )
    assert res.returncode == 0
    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Custom Name ────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_override_click_command(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import click

        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding Command!')

        @click.command(cls=OverrideCommand)
        def cli():
            """My help text"""
            print('Hello, world!')
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_override_click_group(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import click

        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding Command!')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding Command!')

        class OverrideGroup(OverrideCommand, click.Group):
            command_class = OverrideCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding Command!')

        @click.group(cls=OverrideGroup)
        def cli():
            """My help text"""

        @cli.command("subcommand")
        def subcommand():
            """Subcommand help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS] COMMAND [ARGS]...                               \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────────────────────────╮
│ subcommand                        Subcommand help text                                           │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_override_rich_click_command(mock_script_writer: WriteScript) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        # Test if robust to subclassing
        class OverrideRichCommand(click.RichCommand):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        @click.command(cls=OverrideRichCommand)
        def cli():
            """My help text"""
            print('Hello, world!')
        '''
    )

    res = run_as_subprocess([sys.executable, "-m", "rich_click", "mymodule:cli", "--help"])
    assert res.returncode == 0

    assert res.returncode == 0
    assert res.stdout.decode() == snapshot(
        """\
I am overriding RichCommand! (usage)
I am overriding RichCommand! (help_text)
I am overriding RichCommand! (options)
I am overriding RichCommand! (epilog)
"""
    )


def test_override_rich_click_group(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        # Test if robust to subclassing
        class OverrideRichCommand(click.RichCommand):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        class OverrideRichGroup(OverrideRichCommand, click.RichGroup):
            command_class = OverrideRichCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding RichCommand (format_commands)!')

        @click.group(cls=OverrideRichGroup)
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
I am overriding RichCommand! (usage)
I am overriding RichCommand! (help_text)
I am overriding RichCommand! (options)
I am overriding RichCommand! (epilog)
"""
    )


def test_override_rich_click_command_collection(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        # Test if robust to subclassing
        class OverrideRichCommand(click.RichCommand):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        class OverrideRichCommandCollection(OverrideRichCommand, click.RichCommandCollection):
            command_class = OverrideRichCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding RichCommand (format_commands)!')

        @click.group(cls=OverrideRichCommandCollection)
        def cli():
            """My help text"""

        @cli.command("subcommand")
        def subcommand():
            """Subcommand help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
I am overriding RichCommand! (usage)
I am overriding RichCommand! (help_text)
I am overriding RichCommand! (options)
I am overriding RichCommand! (epilog)
"""
    )


def test_override_guard_click_command(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import click

        # Test if robust to subclassing
        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        @click.group(cls=OverrideCommand)
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_override_guard_click_group(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import click

        # Test if robust to subclassing
        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        class OverrideGroup(OverrideCommand, click.Group):
            command_class = OverrideCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding RichCommand (format_commands)!')

        @click.group(cls=OverrideGroup)
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS] COMMAND [ARGS]...                               \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_override_guard_disabled_click_from_rich_click(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        # Test if robust to subclassing
        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        class OverrideGroup(OverrideCommand, click.Group):
            command_class = OverrideCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding RichCommand (format_commands)!')

        @click.group(cls=OverrideGroup)
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--no-patch-rich-click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        "I am overriding RichCommand! (usage)I am overriding RichCommand! (help_text)I am overriding RichCommand! (options)I am overriding RichCommand! (epilog)\n"
    )


def test_override_guard_enabled_click_from_rich_click(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        # Test if robust to subclassing
        class OverrideCommand(click.Command):
            def format_usage(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (usage)')
            def format_help_text(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (help_text)')
            def format_options(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (options)')
            def format_epilog(self, ctx, formatter):
                formatter.write('I am overriding RichCommand! (epilog)')

        class OverrideGroup(OverrideCommand, click.Group):
            command_class = OverrideCommand
            def format_commands(self, ctx, formatter):
                formatter.write('I am overriding RichCommand (format_commands)!')

        @click.group(cls=OverrideGroup)
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS] COMMAND [ARGS]...                               \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )


def test_error_to_stderr(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import click

        @click.group("foo")
        def foo():
            """foo group"""

        @foo.command("bar")
        def bar():
            """bar command"""
        '''
    )

    res_grp = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:foo", "--bad-input"],
    )
    assert res_grp.returncode == 2
    assert res_grp.stdout.decode() == ""
    assert res_grp.stderr.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS] COMMAND [ARGS]...                               \n\
                                                                                                    \n\
 Try 'python -m src.rich_click.mymodule --help' for help                                            \n\
╭─ Error ──────────────────────────────────────────────────────────────────────────────────────────╮
│ No such option: --bad-input                                                                      │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
                                                                                                    \n\
"""
    )

    res_cmd = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "mymodule:foo", "bar", "--bad-input"],
    )
    assert res_cmd.returncode == 2

    assert res_grp.stdout.decode() == ""
    assert res_grp.stderr.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS] COMMAND [ARGS]...                               \n\
                                                                                                    \n\
 Try 'python -m src.rich_click.mymodule --help' for help                                            \n\
╭─ Error ──────────────────────────────────────────────────────────────────────────────────────────╮
│ No such option: --bad-input                                                                      │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
                                                                                                    \n\
"""
    )


rich_version = packaging.version.parse(version("rich"))


@pytest.mark.skipif(rich_version < packaging.version.parse("13.0.0"), reason="Rich <13 renders differently")
def test_cli_output_html(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        @click.command()
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--output", "html", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<style>

body {
    color: #000000;
    background-color: #ffffff;
}
</style>
</head>
<body>
    <pre style="font-family:Menlo,'DejaVu Sans Mono',consolas,'Courier New',monospace"><code style="font-family:inherit">                                                                                                    \n\
 <span style="color: #808000; text-decoration-color: #808000">Usage:</span> <span style="font-weight: bold">python -m src.rich_click.mymodule</span> [<span style="color: #008080; text-decoration-color: #008080; font-weight: bold">OPTIONS</span>]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
<span style="color: #7f7f7f; text-decoration-color: #7f7f7f">╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮</span>
<span style="color: #7f7f7f; text-decoration-color: #7f7f7f">│</span> <span style="color: #008080; text-decoration-color: #008080; font-weight: bold">--help</span>  Show this message and exit.                                                              <span style="color: #7f7f7f; text-decoration-color: #7f7f7f">│</span>
<span style="color: #7f7f7f; text-decoration-color: #7f7f7f">╰──────────────────────────────────────────────────────────────────────────────────────────────────╯</span>
</code></pre>
</body>
</html>
"""
    )


def test_cli_output_svg(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        @click.command()
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--output", "svg", "mymodule:cli", "--help"],
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
<svg class="rich-terminal" viewBox="0 0 1238 245.2" xmlns="http://www.w3.org/2000/svg">
    <!-- Generated with Rich https://www.textualize.io -->
    <style>

    @font-face {
        font-family: "Fira Code";
        src: local("FiraCode-Regular"),
                url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff2/FiraCode-Regular.woff2") format("woff2"),
                url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff/FiraCode-Regular.woff") format("woff");
        font-style: normal;
        font-weight: 400;
    }
    @font-face {
        font-family: "Fira Code";
        src: local("FiraCode-Bold"),
                url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff2/FiraCode-Bold.woff2") format("woff2"),
                url("https://cdnjs.cloudflare.com/ajax/libs/firacode/6.2.0/woff/FiraCode-Bold.woff") format("woff");
        font-style: bold;
        font-weight: 700;
    }

    .terminal-237638219-matrix {
        font-family: Fira Code, monospace;
        font-size: 20px;
        line-height: 24.4px;
        font-variant-east-asian: full-width;
    }

    .terminal-237638219-title {
        font-size: 18px;
        font-weight: bold;
        font-family: arial;
    }

    .terminal-237638219-r1 { fill: #c5c8c6 }
.terminal-237638219-r2 { fill: #d0b344 }
.terminal-237638219-r3 { fill: #c5c8c6;font-weight: bold }
.terminal-237638219-r4 { fill: #68a0b3;font-weight: bold }
.terminal-237638219-r5 { fill: #868887 }
    </style>

    <defs>
    <clipPath id="terminal-237638219-clip-terminal">
      <rect x="0" y="0" width="1219.0" height="194.2" />
    </clipPath>
    <clipPath id="terminal-237638219-line-0">
    <rect x="0" y="1.5" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-1">
    <rect x="0" y="25.9" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-2">
    <rect x="0" y="50.3" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-3">
    <rect x="0" y="74.7" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-4">
    <rect x="0" y="99.1" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-5">
    <rect x="0" y="123.5" width="1220" height="24.65"/>
            </clipPath>
<clipPath id="terminal-237638219-line-6">
    <rect x="0" y="147.9" width="1220" height="24.65"/>
            </clipPath>
    </defs>

    <rect fill="#292929" stroke="rgba(255,255,255,0.35)" stroke-width="1" x="1" y="1" width="1236" height="243.2" rx="8"/><text class="terminal-237638219-title" fill="#c5c8c6" text-anchor="middle" x="618" y="27">mymodule&#160;--help</text>
            <g transform="translate(26,22)">
            <circle cx="0" cy="0" r="7" fill="#ff5f57"/>
            <circle cx="22" cy="0" r="7" fill="#febc2e"/>
            <circle cx="44" cy="0" r="7" fill="#28c840"/>
            </g>
        \n\
    <g transform="translate(9, 41)" clip-path="url(#terminal-237638219-clip-terminal)">
    \n\
    <g class="terminal-237638219-matrix">
    <text class="terminal-237638219-r1" x="1220" y="20" textLength="12.2" clip-path="url(#terminal-237638219-line-0)">
</text><text class="terminal-237638219-r2" x="12.2" y="44.4" textLength="73.2" clip-path="url(#terminal-237638219-line-1)">Usage:</text><text class="terminal-237638219-r3" x="97.6" y="44.4" textLength="402.6" clip-path="url(#terminal-237638219-line-1)">python&#160;-m&#160;src.rich_click.mymodule</text><text class="terminal-237638219-r1" x="512.4" y="44.4" textLength="12.2" clip-path="url(#terminal-237638219-line-1)">[</text><text class="terminal-237638219-r4" x="524.6" y="44.4" textLength="85.4" clip-path="url(#terminal-237638219-line-1)">OPTIONS</text><text class="terminal-237638219-r1" x="610" y="44.4" textLength="12.2" clip-path="url(#terminal-237638219-line-1)">]</text><text class="terminal-237638219-r1" x="1220" y="44.4" textLength="12.2" clip-path="url(#terminal-237638219-line-1)">
</text><text class="terminal-237638219-r1" x="1220" y="68.8" textLength="12.2" clip-path="url(#terminal-237638219-line-2)">
</text><text class="terminal-237638219-r1" x="12.2" y="93.2" textLength="146.4" clip-path="url(#terminal-237638219-line-3)">My&#160;help&#160;text</text><text class="terminal-237638219-r1" x="1220" y="93.2" textLength="12.2" clip-path="url(#terminal-237638219-line-3)">
</text><text class="terminal-237638219-r1" x="1220" y="117.6" textLength="12.2" clip-path="url(#terminal-237638219-line-4)">
</text><text class="terminal-237638219-r5" x="0" y="142" textLength="24.4" clip-path="url(#terminal-237638219-line-5)">╭─</text><text class="terminal-237638219-r5" x="24.4" y="142" textLength="109.8" clip-path="url(#terminal-237638219-line-5)">&#160;Options&#160;</text><text class="terminal-237638219-r5" x="134.2" y="142" textLength="1061.4" clip-path="url(#terminal-237638219-line-5)">───────────────────────────────────────────────────────────────────────────────────────</text><text class="terminal-237638219-r5" x="1195.6" y="142" textLength="24.4" clip-path="url(#terminal-237638219-line-5)">─╮</text><text class="terminal-237638219-r1" x="1220" y="142" textLength="12.2" clip-path="url(#terminal-237638219-line-5)">
</text><text class="terminal-237638219-r5" x="0" y="166.4" textLength="12.2" clip-path="url(#terminal-237638219-line-6)">│</text><text class="terminal-237638219-r4" x="24.4" y="166.4" textLength="73.2" clip-path="url(#terminal-237638219-line-6)">--help</text><text class="terminal-237638219-r1" x="122" y="166.4" textLength="329.4" clip-path="url(#terminal-237638219-line-6)">Show&#160;this&#160;message&#160;and&#160;exit.</text><text class="terminal-237638219-r5" x="1207.8" y="166.4" textLength="12.2" clip-path="url(#terminal-237638219-line-6)">│</text><text class="terminal-237638219-r1" x="1220" y="166.4" textLength="12.2" clip-path="url(#terminal-237638219-line-6)">
</text><text class="terminal-237638219-r5" x="0" y="190.8" textLength="1220" clip-path="url(#terminal-237638219-line-7)">╰──────────────────────────────────────────────────────────────────────────────────────────────────╯</text><text class="terminal-237638219-r1" x="1220" y="190.8" textLength="12.2" clip-path="url(#terminal-237638219-line-7)">
</text>
    </g>
    </g>
</svg>
"""
    )


def test_cli_output_text(mock_script_writer: Callable[[str], Path]) -> None:
    mock_script_writer(
        '''
        import rich_click as click

        @click.command()
        def cli():
            """My help text"""
        ''',
    )

    res = run_as_subprocess(
        [sys.executable, "-m", "src.rich_click", "--output", "text", "mymodule:cli", "--help"],
        # Even with FORCE_COLOR=True, no color should render with --output=text
        env={"FORCE_COLOR": "True"},
    )
    assert res.returncode == 0

    assert res.stdout.decode() == snapshot(
        """\
                                                                                                    \n\
 Usage: python -m src.rich_click.mymodule [OPTIONS]                                                 \n\
                                                                                                    \n\
 My help text                                                                                       \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )



from __future__ import annotations

import warnings

import click


try:
    with warnings.catch_warnings():
        warnings.simplefilter(category=DeprecationWarning, action="ignore")
        click_version = click.__version__
except Exception:
    # Click 9+ deprecated __version__, so all these checks must necessarily be False if __version__ doesn't exist.
    CLICK_IS_BEFORE_VERSION_821 = False
    CLICK_IS_BEFORE_VERSION_82 = False
    CLICK_IS_BEFORE_VERSION_9X = False
    CLICK_IS_VERSION_80 = False
else:
    _major = int(click_version.split(".")[0])  # type: ignore[attr-defined,unused-ignore]
    _minor = int(click_version.split(".")[1])  # type: ignore[attr-defined,unused-ignore]
    _patch = int(click_version.split(".")[2])  # type: ignore[attr-defined,unused-ignore]

    CLICK_IS_BEFORE_VERSION_821 = (_major, _minor, _patch) < (8, 2, 1)
    CLICK_IS_BEFORE_VERSION_82 = (_major, _minor) < (8, 2)
    CLICK_IS_BEFORE_VERSION_9X = _major < 9
    CLICK_IS_VERSION_80 = (_major, _minor) == (8, 0)



# ruff: noqa: D101,D103,D401,E501
import builtins
import importlib
from pathlib import Path
from typing import Any, List

import pytest
from click.testing import CliRunner
from inline_snapshot import snapshot

import rich_click
from tests.conftest import WriteScript


# These tests are to assert various optimizations.
# There are certain imports that significantly slow down execution:
#
# - Importing anything from `rich` during execution is unnecessary and should not occur.
# - Importing `markdown_it` except when necessary significantly slows down help text rendering.
# - Calling `importlib.metadata.version()` is slow. Click has made a decision to deprecate `__version__`,
#   and so the latest versions of click call `importlib.metadata.version()` when you access
#   `click.__version__`, which slows things down. During execution, we want to avoid accessing
#   `click.__version__`. A slightly lazy albeit acceptable way to assert that
#   `importlib.metadata.version()` is never called is to just assert that `importlib.metadata`
#   is not imported at all. During --help, however, we accept that this call will be made.


@pytest.fixture
def check_imports_cli(mock_script_writer: WriteScript) -> Path:
    return mock_script_writer(
        """
        import rich_click as click

        @click.command()
        def cli():
            print("Hello, world!")

        if __name__ == "__main__":
            cli()
        """,
        module_name="my_script.py",
    )


@pytest.fixture
def recorded_imports(monkeypatch: pytest.MonkeyPatch) -> List[str]:
    modules: List[str] = []

    _import = builtins.__import__

    def noisy_import(name: str, *args: Any, **kwargs: Any) -> Any:
        modules.append(name)
        return _import(name, *args, **kwargs)

    monkeypatch.setattr(builtins, "__import__", noisy_import)
    return modules


def test_imports_during_execution(recorded_imports: List[str], cli_runner: CliRunner) -> None:
    importlib.reload(rich_click)

    @rich_click.command()
    def cli() -> None:
        print("Hello, world!")

    res = cli_runner.invoke(cli)
    assert res.exit_code == 0
    assert res.stdout == "Hello, world!\n"

    assert any(m.startswith("click.") or m == "click" for m in recorded_imports)
    assert not any(m.startswith("rich.") or m == "rich" for m in recorded_imports)
    assert not any(m.startswith("markdown_it.") or m == "markdown_it" for m in recorded_imports)
    assert not any(m.startswith("importlib.") or m == "importlib" for m in recorded_imports)


def test_imports_during_help(recorded_imports: List[str], cli_runner: CliRunner) -> None:
    importlib.reload(rich_click)

    @rich_click.command()
    def cli() -> None:
        print("Hello, world!")

    res = cli_runner.invoke(cli, "--help")
    assert res.exit_code == 0
    assert res.stdout == snapshot(
        """\
                                                                                                    \n\
 Usage: cli [OPTIONS]                                                                               \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )

    assert any(m.startswith("click.") or m == "click" for m in recorded_imports)
    assert any(m.startswith("rich.") or m == "rich" for m in recorded_imports)
    assert not any(m.startswith("markdown_it.") or m == "markdown_it" for m in recorded_imports)


def test_imports_during_execution_rich_click_cli(
    recorded_imports: List[str],
    cli_runner: CliRunner,
    check_imports_cli: Path,
) -> None:
    importlib.reload(rich_click)
    from rich_click.cli import main

    res = cli_runner.invoke(main, str(check_imports_cli / "my_script.py"))
    assert res.exit_code == 0
    assert res.stdout == "Hello, world!\n"

    assert any(m.startswith("click.") or m == "click" for m in recorded_imports)
    assert not any(m.startswith("rich.") or m == "rich" for m in recorded_imports)
    assert not any(m.startswith("markdown_it.") or m == "markdown_it" for m in recorded_imports)


def test_imports_during_help_rich_click_cli(
    recorded_imports: List[str],
    cli_runner: CliRunner,
    check_imports_cli: Path,
) -> None:
    importlib.reload(rich_click)
    from rich_click.cli import main

    res = cli_runner.invoke(main, f"{check_imports_cli / 'my_script.py'} --help")
    assert res.exit_code == 0
    assert res.stdout == snapshot(
        """\
                                                                                                    \n\
 Usage: mymodule [OPTIONS]                                                                          \n\
                                                                                                    \n\
╭─ Options ────────────────────────────────────────────────────────────────────────────────────────╮
│ --help  Show this message and exit.                                                              │
╰──────────────────────────────────────────────────────────────────────────────────────────────────╯
"""
    )

    assert any(m.startswith("click.") or m == "click" for m in recorded_imports)
    assert any(m.startswith("rich.") or m == "rich" for m in recorded_imports)
    assert not any(m.startswith("markdown_it.") or m == "markdown_it" for m in recorded_imports)



# fmt: off
from typing import Union

import rich.box
from rich.box import Box


HORIZONTALS_TOP: Box = Box(
    " ── \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
)

HORIZONTALS_DOUBLE_TOP: Box = Box(
    " ══ \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
)

BLANK: Box = Box(
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
)

BLANK.top = ""
BLANK.top_left = ""
BLANK.top_right = " " * 800  # Reasonably ensure padding
BLANK.top_divider = ""

def get_box(box: Union[str, Box]) -> Box:
    """Retrieve a Rich Box by name."""
    if isinstance(box, Box):
        return box
    if box == box.upper() and box in globals():
        return globals()[box]  # type: ignore[no-any-return]
    return getattr(rich.box, box)  # type: ignore[no-any-return]



from __future__ import annotations

import errno
import os
import sys
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    TextIO,
    Type,
    Union,
    cast,
    overload,
)

import click

# Group, Command, and CommandCollection need to be imported directly,
# or else rich_click.cli.patch() causes a recursion error.
from click import Command, CommandCollection, Group
from click.utils import PacifyFlushWrapper

from rich_click.rich_context import RichContext
from rich_click.rich_help_configuration import RichHelpConfiguration
from rich_click.rich_help_formatter import RichHelpFormatter


if TYPE_CHECKING:  # pragma: no cover
    from rich.console import Console

    from rich_click.rich_help_rendering import RichPanelRow
    from rich_click.rich_panel import RichCommandPanel, RichPanel


# TLDR: if a subcommand overrides one of the methods called by `RichCommand.format_help`,
# then the text won't render properly. The fix is to not rely on the composability of the API,
# and to instead force everything to use RichCommand's methods.
OVERRIDES_GUARD: bool = False


class RichCommand(Command):
    """
    Richly formatted click Command.

    Inherits click.Command and overrides help and error methods
    to print richly formatted output.

    This class can be used as a mixin for other click command objects.
    """

    context_class: Type[RichContext] = RichContext
    _formatter: Optional[RichHelpFormatter] = None

    def __init__(
        self,
        *args: Any,
        aliases: Optional[Iterable[str]] = None,
        panels: Optional[List["RichPanel[Any, Any]"]] = None,
        panel: Optional[Union[str, List[str]]] = None,
        **kwargs: Any,
    ) -> None:
        """Create Rich Command instance."""
        super().__init__(*args, **kwargs)
        self.panel = panel
        self.panels: List["RichPanel[Any, Any]"] = panels or []
        self.aliases: Iterable[str] = aliases or []
        if not hasattr(self, "_help_option"):
            self._help_option = None

    @property
    def console(self) -> Optional["Console"]:
        """
        Rich Console.

        This is a separate instance from the help formatter that allows full control of the
        console configuration.

        See `rich_config` decorator for how to apply the settings.
        """
        warnings.warn(
            "RichCommand.console is deprecated. Please use the click.Context's console instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.context_settings.get("rich_console")

    def to_info_dict(self, ctx: click.Context) -> Dict[str, Any]:
        info = super().to_info_dict(ctx)
        info["panels"] = [p.to_info_dict(ctx) for p in self.panels]
        info["aliases"] = list(self.aliases) if self.aliases is not None else None
        return info

    @property
    def help_config(self) -> Optional[RichHelpConfiguration]:
        """Rich Help Configuration."""
        warnings.warn(
            "RichCommand.help_config is deprecated. Please use the click.Context's help config instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        cfg = self.context_settings.get("rich_help_config")
        if isinstance(cfg, Mapping):
            return RichHelpConfiguration(**cfg)
        return cfg

    def _generate_rich_help_config(self) -> RichHelpConfiguration:
        """
        Use for error handling when a Context is not available.

        If the Context is available, then the help configuration in the Context
        should be preferred.
        """
        cfg = self.context_settings.get("rich_help_config", {})
        try:
            if isinstance(cfg, Mapping):
                return RichHelpConfiguration.load_from_globals(**cfg)
            elif isinstance(cfg, RichHelpConfiguration):
                return cfg
        except Exception as e:
            click.echo(f"{e.__class__.__name__}{e.args}", file=sys.stderr)
        return RichHelpConfiguration()

    def _error_formatter(self) -> RichHelpFormatter:
        from click import get_current_context

        def _get_formatter() -> RichHelpFormatter:
            config = self._generate_rich_help_config()
            formatter = self.context_class.formatter_class(
                config=config,
                export_console_as=(
                    self.context_class.export_console_as if self.context_class.errors_in_output_format else None
                ),
            )
            return formatter

        try:
            ctx: RichContext = get_current_context()  # type: ignore[assignment]
        except RuntimeError:
            formatter = _get_formatter()
        else:
            if not isinstance(ctx, RichContext):
                formatter = _get_formatter()
            else:
                formatter = ctx.make_formatter(error_mode=True)
        return formatter

    @overload
    def main(
        self,
        args: Sequence[str] | None = None,
        prog_name: str | None = None,
        complete_var: str | None = None,
        standalone_mode: Literal[True] = True,
        **extra: Any,
    ) -> NoReturn: ...

    @overload
    def main(
        self,
        args: Sequence[str] | None = None,
        prog_name: str | None = None,
        complete_var: str | None = None,
        standalone_mode: bool = ...,
        **extra: Any,
    ) -> Any: ...

    def main(
        self,
        args: Optional[Sequence[str]] = None,
        prog_name: Optional[str] = None,
        complete_var: Optional[str] = None,
        standalone_mode: bool = True,
        windows_expand_args: bool = True,
        **extra: Any,
    ) -> Any:
        # It's not feasible to use super().main() in this context and retain exact parity in behavior.
        # The reason why is explained in a comment in click's source code in the "except Exit as e" block.

        if args is None:
            args = sys.argv[1:]

            if os.name == "nt" and windows_expand_args:
                from click.utils import _expand_args

                args = _expand_args(args)
        else:
            args = list(args)

        if TYPE_CHECKING:  # pragma: no cover
            assert args is not None

        if prog_name is None:
            from click.utils import _detect_program_name

            prog_name = _detect_program_name()

        # Process shell completion requests and exit early.
        self._main_shell_completion(extra, prog_name, complete_var)

        ctx = None

        try:
            try:
                with self.make_context(prog_name, args, **extra) as ctx:
                    rv = self.invoke(ctx)
                    if not standalone_mode:
                        return rv
                    # it's not safe to `ctx.exit(rv)` here!
                    # note that `rv` may actually contain data like "1" which
                    # has obvious effects
                    # more subtle case: `rv=[None, None]` can come out of
                    # chained commands which all returned `None` -- so it's not
                    # even always obvious that `rv` indicates success/failure
                    # by its truthiness/falsiness
                    ctx.exit()
            except (EOFError, KeyboardInterrupt):
                click.echo(file=sys.stderr)
                raise click.exceptions.Abort() from None
            except click.exceptions.ClickException as e:
                from rich_click._compat_click import CLICK_IS_BEFORE_VERSION_82

                if not CLICK_IS_BEFORE_VERSION_82:
                    # `except click.exceptions.NoArgsIsHelpError as e:` breaks for click<8.2.
                    if isinstance(e, click.exceptions.NoArgsIsHelpError):  #
                        print(e.message)
                        sys.exit(e.exit_code)
                if not standalone_mode:
                    raise
                formatter = self._error_formatter()
                formatter.write_error(e)
                print(formatter.getvalue(), file=sys.stderr, end="")
                sys.exit(e.exit_code)
            except OSError as e:
                if e.errno == errno.EPIPE:
                    sys.stdout = cast(TextIO, PacifyFlushWrapper(sys.stdout))
                    sys.stderr = cast(TextIO, PacifyFlushWrapper(sys.stderr))
                    sys.exit(1)
                else:
                    raise
        except click.exceptions.Exit as e:
            if standalone_mode:
                sys.exit(e.exit_code)
            else:
                return e.exit_code
        except click.exceptions.Abort:
            if not standalone_mode:
                raise
            try:
                formatter = self._error_formatter()
            except Exception:
                click.echo("Aborted!", file=sys.stderr)
            else:
                formatter.write_abort()
                print(formatter.getvalue(), file=sys.stderr, end="")
            finally:
                sys.exit(1)

    # Mypy complains about Liskov substitution principle violations.
    # We opt to ignore mypy here.

    def format_help(self, ctx: RichContext, formatter: RichHelpFormatter) -> None:  # type: ignore[override]
        if OVERRIDES_GUARD:
            prevent_incompatible_overrides(self, "RichCommand", ctx, formatter)
        else:
            self.format_usage(ctx, formatter)
            self.format_help_text(ctx, formatter)
            self.format_options(ctx, formatter)
            self.format_epilog(ctx, formatter)

    def format_help_text(self, ctx: RichContext, formatter: RichHelpFormatter) -> None:  # type: ignore[override]
        from rich_click.rich_help_rendering import get_rich_help_text

        get_rich_help_text(self, ctx, formatter)

    # TODO:
    #  Switching from base click to rich click causes mypy problems.
    #  Either we: (a) swap MRO (incompatible with click 9, without handling 8 and 9 differently)
    #  or (b) we allow issues when users attempt multiple inheritance with a RichCommand
    #  or (c) we use incorrect types here.
    #  We are looking for a solution that fixes all 3. For now, we opt for (c).
    def format_options(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        from rich.table import Table

        from rich_click.rich_panel import construct_panels

        panels = construct_panels(self, ctx, formatter)  # type: ignore[arg-type]
        for panel in panels:
            p = panel.render(self, ctx, formatter)  # type: ignore[arg-type]
            if not isinstance(p.renderable, Table) or len(p.renderable.rows) > 0:
                formatter.write(p)  # type: ignore[arg-type]

    def format_epilog(self, ctx: RichContext, formatter: RichHelpFormatter) -> None:  # type: ignore[override]
        from rich_click.rich_help_rendering import get_rich_epilog

        get_rich_epilog(self, ctx, formatter)

    def get_help_option(self, ctx: click.Context) -> Union[click.Option, None]:
        """
        Return the help option object.

        Skipped if :attr:`add_help_option` is ``False``.

        .. versionchanged:: 8.1.8
            The help option is now cached to avoid creating it multiple times.
        """
        help_option_names = self.get_help_option_names(ctx)

        if not help_option_names or not self.add_help_option:
            return None

        # Cache the help option object in private _help_option attribute to
        # avoid creating it multiple times. Not doing this will break the
        # callback ordering by iter_params_for_processing(), which relies on
        # object comparison.
        if self._help_option is None:
            # Avoid circular import.
            from rich_click.decorators import help_option

            # Apply help_option decorator and pop resulting option
            help_option(*help_option_names)(self)
            self._help_option = self.params.pop()  # type: ignore[assignment]

        return self._help_option

    def get_rich_table_row(
        self,
        ctx: "RichContext",
        formatter: "RichHelpFormatter",
        panel: Optional["RichCommandPanel"] = None,
    ) -> "RichPanelRow":
        """Create a row for the rich table corresponding with this parameter."""
        from rich_click.rich_help_rendering import get_command_rich_table_row

        return get_command_rich_table_row(self, ctx, formatter, panel)

    def add_panel(self, panel: "RichPanel[Any, Any]") -> None:
        """Add a RichPanel to the RichCommand."""
        self.panels.append(panel)


class RichGroup(RichCommand, Group):
    """
    Richly formatted click Group.

    Inherits click.Group and overrides help and error methods
    to print richly formatted output.
    """

    command_class: Optional[Type[RichCommand]] = RichCommand
    group_class: Optional[Union[Type[Group], Type[type]]] = type

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create RichGroup instance."""
        super().__init__(*args, **kwargs)

        self._alias_mapping: Dict[str, str] = {}
        # This allows non-RichCommands to be assigned to panels
        # + assigns without requiring mutation of panels.
        self._panel_command_mapping: Dict[str, List[str]] = {}

        for name in self.commands:
            cmd = self.commands[name]

            aliases: Optional[Iterable[str]] = getattr(cmd, "aliases", None)
            if aliases:
                for alias in aliases:
                    self._alias_mapping[alias] = name

            panel: Optional[str] = getattr(cmd, "panel", None)
            if cmd.name and panel:
                self.add_command_to_panel(cmd, panel)

    def format_commands(self, ctx: click.Context, formatter: click.HelpFormatter) -> None:
        # Not used
        pass

    def format_help(self, ctx: RichContext, formatter: RichHelpFormatter) -> None:  # type: ignore[override]
        if OVERRIDES_GUARD:
            prevent_incompatible_overrides(self, "RichGroup", ctx, formatter)
        else:
            self.format_usage(ctx, formatter)
            self.format_help_text(ctx, formatter)
            self.format_options(ctx, formatter)
            self.format_epilog(ctx, formatter)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Alias for :meth:`main`."""
        # Include this here because I run into a false warning
        # in the PyCharm IDE otherwise; for some reason PyCharm doesn't
        # seem to think RichGroups are callable. (No issues with Mypy, though.)
        return super().__call__(*args, **kwargs)

    @overload
    def command(self, __func: Callable[..., Any]) -> RichCommand: ...

    @overload
    def command(self, *args: Any, **kwargs: Any) -> Callable[[Callable[..., Any]], RichCommand]: ...

    def command(self, *args: Any, **kwargs: Any) -> Union[Callable[[Callable[..., Any]], RichCommand], RichCommand]:
        """
        A shortcut decorator for declaring and attaching a command to
        the group. This takes the same arguments as :func:`command` and
        immediately registers the created command with this group by
        calling :meth:`add_command`.

        To customize the command class used, set the
        :attr:`command_class` attribute.

        .. versionchanged:: 8.1
            This decorator can be applied without parentheses.

        .. versionchanged:: 8.0
            Added the :attr:`command_class` attribute.
        """  # noqa: D401
        from rich_click.decorators import command

        func: Optional[Callable[..., Any]] = None

        if args and callable(args[0]):
            assert len(args) == 1 and not kwargs, "Use 'command(**kwargs)(callable)' to provide arguments."
            (func,) = args
            args = ()

        cls: Optional[Type[Command]] = kwargs.get("cls")
        if self.command_class and cls is None:
            kwargs["cls"] = cls = self.command_class

        def decorator(f: Callable[..., Any]) -> RichCommand:
            if cls and not issubclass(cls, RichCommand):
                panel = kwargs.pop("panel", None)
                aliases = kwargs.pop("aliases", None)
            else:
                panel = kwargs.get("panel")
                aliases = kwargs.get("aliases")
            cmd: RichCommand = command(*args, **kwargs)(f)
            self.add_command(cmd)
            self._handle_extras_add_command(cmd, aliases=aliases, panel=panel)
            return cmd

        if func is not None:
            return decorator(func)

        return decorator

    @overload
    def group(self, __func: Callable[..., Any]) -> "RichGroup": ...

    @overload
    def group(self, *args: Any, **kwargs: Any) -> Callable[[Callable[..., Any]], "RichGroup"]: ...

    def group(self, *args: Any, **kwargs: Any) -> Union[Callable[[Callable[..., Any]], "RichGroup"], "RichGroup"]:
        """
        A shortcut decorator for declaring and attaching a group to
        the group. This takes the same arguments as :func:`group` and
        immediately registers the created group with this group by
        calling :meth:`add_command`.

        To customize the group class used, set the :attr:`group_class`
        attribute.

        .. versionchanged:: 8.1
            This decorator can be applied without parentheses.

        .. versionchanged:: 8.0
            Added the :attr:`group_class` attribute.
        """  # noqa: D401
        from rich_click.decorators import group

        func: Optional[Callable[..., Any]] = None

        if args and callable(args[0]):
            assert len(args) == 1 and not kwargs, "Use 'group(**kwargs)(callable)' to provide arguments."
            (func,) = args
            args = ()

        cls: Optional[Union[Type[Group], Type[type]]] = kwargs.get("cls")
        if self.group_class is not None and cls is None:
            if self.group_class is type:
                kwargs["cls"] = cls = type(self)
            else:
                kwargs["cls"] = cls = self.group_class

        def decorator(f: Callable[..., Any]) -> RichGroup:
            if cls and not issubclass(cls, RichCommand):
                panel = kwargs.pop("panel", None)
                aliases = kwargs.pop("aliases", None)
            else:
                panel = kwargs.get("panel")
                aliases = kwargs.get("aliases")
            cmd: RichGroup = group(*args, **kwargs)(f)
            self.add_command(cmd)
            self._handle_extras_add_command(cmd, aliases=aliases, panel=panel)
            return cmd

        if func is not None:
            return decorator(func)

        return decorator

    def _handle_extras_add_command(
        self,
        cmd: click.Command,
        name: Optional[str] = None,
        aliases: Optional[Iterable[str]] = None,
        panel: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """
        Create backwards compatibility with add_command() subclass interfaces
        that have not migrated to rich-click's 1.9.0 add_command(...).

        This should stay in place until a 2.0 release.

        In the meanwhile, devs should get mypy errors indicating that add_command()
        does not implement all the proper kwargs.
        """
        _name: str = name or cmd.name  # type: ignore[assignment]
        if aliases:
            for alias in aliases:
                self._alias_mapping[alias] = _name
        additional_aliases = getattr(cmd, "aliases", None)
        if additional_aliases:
            for alias in additional_aliases:
                self._alias_mapping[alias] = _name
        self._alias_mapping.pop(_name, None)
        panel = panel or getattr(cmd, "panel", None)
        if panel:
            self.add_command_to_panel(cmd, panel)

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        _cmd_name = self._alias_mapping.get(cmd_name, cmd_name)
        return super().get_command(ctx, _cmd_name)

    def add_command(
        self,
        cmd: click.Command,
        name: Optional[str] = None,
        aliases: Optional[Iterable[str]] = None,
        panel: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """
        Register another :class:`Command` with this group. If the name
        is not provided, the name of the command is used.
        """
        super().add_command(cmd, name)
        _name: str = name or cmd.name  # type: ignore[assignment]
        if aliases or panel:
            self._handle_extras_add_command(cmd, name=name, aliases=aliases, panel=panel)

    def add_command_to_panel(
        self,
        command: click.Command,
        panel_name: Union[str, Iterable[str]],
    ) -> None:
        if not command.name:
            return
        self._panel_command_mapping.setdefault(command.name, [])
        if isinstance(panel_name, str):
            self._panel_command_mapping[command.name].append(panel_name)
        else:
            self._panel_command_mapping[command.name].extend(panel_name)


RichMultiCommand = RichGroup


class RichCommandCollection(CommandCollection, RichGroup):
    """
    Richly formatted click CommandCollection.

    Inherits click.CommandCollection and overrides help and error methods
    to print richly formatted output.
    """

    def format_help(self, ctx: RichContext, formatter: RichHelpFormatter) -> None:  # type: ignore[override]
        if OVERRIDES_GUARD:
            prevent_incompatible_overrides(self, "RichCommandCollection", ctx, formatter)
        else:
            self.format_usage(ctx, formatter)
            self.format_help_text(ctx, formatter)
            self.format_options(ctx, formatter)
            self.format_epilog(ctx, formatter)


def prevent_incompatible_overrides(
    cmd: RichCommand, class_name: str, ctx: RichContext, formatter: RichHelpFormatter
) -> None:
    """For use by the rich-click CLI."""
    import rich_click.patch
    from rich_click.utils import method_is_from_subclass_of

    cls: Type[RichCommand] = getattr(rich_click.patch, f"_Patched{class_name}")

    for method_name in ["format_usage", "format_help_text", "format_options", "format_epilog"]:
        if method_is_from_subclass_of(cmd.__class__, cls, method_name):
            getattr(RichCommand, method_name)(cmd, ctx, formatter)
        else:
            getattr(cmd, method_name)(ctx, formatter)
